import unittest
from schemaparser import SchemaParser

class TestPrimitiveTypes(unittest.TestCase):
    def test_basic_bool_schema(self):
        DO_NOT_INCLUDE = "doNotInclude"
        DO_NOT_INCLUDE_DESCRIPTION = "Include flag of the layout subset."

        schema = {
            "definitions": {
                DO_NOT_INCLUDE: {
                    "type": "boolean",
                    "description": DO_NOT_INCLUDE_DESCRIPTION
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(DO_NOT_INCLUDE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[DO_NOT_INCLUDE].type, 'boolean')
        self.assertEqual(parser.primitive_type_schemas[DO_NOT_INCLUDE].description, DO_NOT_INCLUDE_DESCRIPTION)

        self.assertIsNone(parser.primitive_type_schemas[DO_NOT_INCLUDE].title)
        self.assertIsNone(parser.primitive_type_schemas[DO_NOT_INCLUDE].default)
        self.assertIsNone(parser.primitive_type_schemas[DO_NOT_INCLUDE].examples)

    def test_full_bool_schema(self):
        DO_NOT_INCLUDE = "doNotInclude"
        DO_NOT_INCLUDE_DESCRIPTION = "Include flag of the layout subset."

        schema = {
            "definitions": {
                DO_NOT_INCLUDE: {
                    "title": "Title example",
                    "type": "boolean",
                    "description": DO_NOT_INCLUDE_DESCRIPTION,
                    "default": False,
                    "examples": [
                        True,
                        False
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(DO_NOT_INCLUDE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[DO_NOT_INCLUDE].type, 'boolean')
        self.assertEqual(parser.primitive_type_schemas[DO_NOT_INCLUDE].description, DO_NOT_INCLUDE_DESCRIPTION)
        self.assertEqual(parser.primitive_type_schemas[DO_NOT_INCLUDE].title, 'Title example')
        self.assertEqual(parser.primitive_type_schemas[DO_NOT_INCLUDE].default, False)
        self.assertEqual(parser.primitive_type_schemas[DO_NOT_INCLUDE].examples, [True, False])
        

    def test_basic_string_schema(self):
        MESSAGE = "message"
        MESSAGE_DESCRIPTION= "The error message."

        schema = {
            "definitions": {
                MESSAGE: {
                    "type": "string",
                    "description": MESSAGE_DESCRIPTION
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(MESSAGE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[MESSAGE].type, 'string')
        self.assertEqual(parser.primitive_type_schemas[MESSAGE].description, MESSAGE_DESCRIPTION)

        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].min_length)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].max_length)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].pattern)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].values)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].title)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].default)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].examples)

    def test_full_string_schema(self):
        MESSAGE = "message"
        MESSAGE_DESCRIPTION = "The error message."
        MESSAGE_PATTERN = "^[0-9A-Za-z ]+\\.$"

        schema = {
            "definitions": {
                MESSAGE: {
                    "type": "string",
                    "description": MESSAGE_DESCRIPTION,
                    "minLength": 2,
                    "maxLength": 3,
                    "pattern": MESSAGE_PATTERN,
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(MESSAGE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[MESSAGE].type, 'string')
        self.assertEqual(parser.primitive_type_schemas[MESSAGE].description, MESSAGE_DESCRIPTION)

        self.assertEqual(parser.primitive_type_schemas[MESSAGE].min_length, 2)
        self.assertEqual(parser.primitive_type_schemas[MESSAGE].max_length, 3)
        self.assertEqual(parser.primitive_type_schemas[MESSAGE].pattern, MESSAGE_PATTERN )

        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].values)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].title)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].default)
        self.assertIsNone(parser.primitive_type_schemas[MESSAGE].examples)

    def test_basic_integer_schema(self):
        CODE = "code"
        CODE_DESCRIPTION = "The error code."

        schema = {
            "definitions": {
                CODE: {
                    "type": "integer",
                    "description": CODE_DESCRIPTION
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(CODE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[CODE].type, 'integer')
        self.assertEqual(parser.primitive_type_schemas[CODE].description, CODE_DESCRIPTION)

        self.assertIsNone(parser.primitive_type_schemas[CODE].multiple_of)
        self.assertIsNone(parser.primitive_type_schemas[CODE].minimum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].exclusive_minimum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].maximum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].exclusive_maximum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].values)
        self.assertIsNone(parser.primitive_type_schemas[CODE].title)
        self.assertIsNone(parser.primitive_type_schemas[CODE].default)
        self.assertIsNone(parser.primitive_type_schemas[CODE].examples)

    def test_full_integer_schema(self):
        CODE = "code"
        CODE_DESCRIPTION = "The error code."

        schema = {
            "definitions": {
                CODE: {
                    "type": "integer",
                    "description": CODE_DESCRIPTION,
                    "multipleOf": 2,
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "maximum": 10,
                    "exclusiveMaximum": False
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(CODE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[CODE].type, 'integer')
        self.assertEqual(parser.primitive_type_schemas[CODE].description, CODE_DESCRIPTION)

        self.assertEqual(parser.primitive_type_schemas[CODE].multiple_of, 2)
        self.assertEqual(parser.primitive_type_schemas[CODE].minimum, 0)
        self.assertEqual(parser.primitive_type_schemas[CODE].exclusive_minimum, True)
        self.assertEqual(parser.primitive_type_schemas[CODE].maximum, 10)
        self.assertEqual(parser.primitive_type_schemas[CODE].exclusive_maximum, False)

        self.assertIsNone(parser.primitive_type_schemas[CODE].values)
        self.assertIsNone(parser.primitive_type_schemas[CODE].title)
        self.assertIsNone(parser.primitive_type_schemas[CODE].default)
        self.assertIsNone(parser.primitive_type_schemas[CODE].examples)
    
    def test_basic_number_schema(self):
        CODE = "code"
        CODE_DESCRIPTION = "The error code."

        schema = {
            "definitions": {
                CODE: {
                    "type": "number",
                    "description": CODE_DESCRIPTION
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(CODE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[CODE].type, 'number')
        self.assertEqual(parser.primitive_type_schemas[CODE].description, CODE_DESCRIPTION)
        
        self.assertIsNone(parser.primitive_type_schemas[CODE].multiple_of)
        self.assertIsNone(parser.primitive_type_schemas[CODE].minimum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].exclusive_minimum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].maximum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].exclusive_maximum)
        self.assertIsNone(parser.primitive_type_schemas[CODE].values)
        self.assertIsNone(parser.primitive_type_schemas[CODE].title)
        self.assertIsNone(parser.primitive_type_schemas[CODE].default)
        self.assertIsNone(parser.primitive_type_schemas[CODE].examples)

    def test_full_number_schema(self):
        CODE = "code"
        CODE_DESCRIPTION = "The error code."

        schema = {
            "definitions": {
                CODE: {
                    "type": "number",
                    "description": CODE_DESCRIPTION,
                    "multipleOf": 20.0,
                    "minimum": -20,
                    "exclusiveMinimum": False,
                    "maximum": 220.00,
                    "exclusiveMaximum": True
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(CODE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[CODE].type, 'number')
        self.assertEqual(parser.primitive_type_schemas[CODE].description, CODE_DESCRIPTION)
        
        self.assertEqual(parser.primitive_type_schemas[CODE].multiple_of, 20)
        self.assertEqual(parser.primitive_type_schemas[CODE].minimum, -20)
        self.assertEqual(parser.primitive_type_schemas[CODE].exclusive_minimum, False)
        self.assertEqual(parser.primitive_type_schemas[CODE].maximum, 220.0)
        self.assertEqual(parser.primitive_type_schemas[CODE].exclusive_maximum, True)

        self.assertIsNone(parser.primitive_type_schemas[CODE].values)
        self.assertIsNone(parser.primitive_type_schemas[CODE].title)
        self.assertIsNone(parser.primitive_type_schemas[CODE].default)
        self.assertIsNone(parser.primitive_type_schemas[CODE].examples)

    def test_guid(self):
        GUID = "guid"
        GUID_DESCRIPTION = "A Globally Unique Identifier (or Universally Unique Identifier) in its string representation as defined in RFC 4122."
        GUID_PATTERN = "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"

        schema = {
            "definitions": {
                GUID: {
                    "type": "string",
                    "description": GUID_DESCRIPTION,
                    "format": "uuid",
                    "pattern": GUID_PATTERN
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(GUID, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[GUID].type, 'UUID')
        self.assertEqual(parser.primitive_type_schemas[GUID].description, GUID_DESCRIPTION)
        self.assertEqual(parser.primitive_type_schemas[GUID].pattern, GUID_PATTERN)

        self.assertIsNone(parser.primitive_type_schemas[GUID].min_length)
        self.assertIsNone(parser.primitive_type_schemas[GUID].max_length)
        self.assertIsNone(parser.primitive_type_schemas[GUID].values)
        self.assertIsNone(parser.primitive_type_schemas[GUID].title)
        self.assertIsNone(parser.primitive_type_schemas[GUID].default)
        self.assertIsNone(parser.primitive_type_schemas[GUID].examples)

    def test_string_enum(self):
        ELEMENT_TYPE = "elementType"
        ELEMENT_TYPE_DESCRIPTION = "The type of the element."

        schema = {
            "definitions": {
                ELEMENT_TYPE: {
                    "type": "string",
                    "description": ELEMENT_TYPE_DESCRIPTION,
                    "enum": [
                        "Wall",
                        "Window",
                        "Object",
                        "Beam"
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)
        
        self.assertIn(ELEMENT_TYPE, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[ELEMENT_TYPE].description, ELEMENT_TYPE_DESCRIPTION)

        self.assertEqual(4, len(parser.primitive_type_schemas[ELEMENT_TYPE].values))
        self.assertIn('Wall', parser.primitive_type_schemas[ELEMENT_TYPE].values)
        self.assertIn('Window', parser.primitive_type_schemas[ELEMENT_TYPE].values)
        self.assertIn('Object', parser.primitive_type_schemas[ELEMENT_TYPE].values)
        self.assertNotIn('Door', parser.primitive_type_schemas[ELEMENT_TYPE].values)
        self.assertFalse('Beam' not in parser.primitive_type_schemas[ELEMENT_TYPE].values)
        
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE].min_length)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE].max_length)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE].pattern)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE].title)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE].default)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE].examples)

    def test_integer_enum(self):
        ELEMENT_TYPE_ID = "elementTypeId"
        ELEMENT_TYPE_ID_DESCRIPTION = "The id of the element type."

        schema = {
            "definitions": {
                ELEMENT_TYPE_ID: {
                    "type": "integer",
                    "description": ELEMENT_TYPE_ID_DESCRIPTION,
                    "enum": [
                        1,
                        3,
                        4,
                        5,
                        8
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)
        
        self.assertIn(ELEMENT_TYPE_ID, parser.primitive_type_schemas)
        
        self.assertEqual(parser.primitive_type_schemas[ELEMENT_TYPE_ID].type, 'integer')
        self.assertEqual(parser.primitive_type_schemas[ELEMENT_TYPE_ID].description, ELEMENT_TYPE_ID_DESCRIPTION)

        self.assertEqual(5, len(parser.primitive_type_schemas[ELEMENT_TYPE_ID].values))
        self.assertIn(1, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertIn(4, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertIn(5, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertNotIn(2, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertNotIn(7, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertFalse(8 not in parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)

        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].multiple_of)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].minimum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].exclusive_minimum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].maximum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].exclusive_maximum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].title)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].default)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].examples)

    def test_number_enum(self):
        ELEMENT_TYPE_ID = "elementTypeId"
        ELEMENT_TYPE_ID_DESCRIPTION = "The id of the element type."

        schema = {
            "definitions": {
                ELEMENT_TYPE_ID: {
                    "type": "number",
                    "description": ELEMENT_TYPE_ID_DESCRIPTION,
                    "enum": [
                        1,
                        3.0,
                        8.2222
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)
        
        self.assertIn(ELEMENT_TYPE_ID, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[ELEMENT_TYPE_ID].description, ELEMENT_TYPE_ID_DESCRIPTION)

        self.assertEqual(3, len(parser.primitive_type_schemas[ELEMENT_TYPE_ID].values))
        self.assertIn(1, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertIn(1.0, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertIn(3, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertIn(3.0, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertNotIn(8, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertNotIn(2, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertNotIn(4.123, parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)
        self.assertFalse(8.2222 not in parser.primitive_type_schemas[ELEMENT_TYPE_ID].values)

        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].multiple_of)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].minimum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].exclusive_minimum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].maximum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].exclusive_maximum)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].title)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].default)
        self.assertIsNone(parser.primitive_type_schemas[ELEMENT_TYPE_ID].examples)


class TestObjects(unittest.TestCase):
    def test_object(self):
        ERROR = "error"
        ERROR_DESCRIPTION = "Error details."

        schema = {
            "definitions": {
                ERROR: {
                    "title": "Object title.",
                    "type": "object",
                    "description": ERROR_DESCRIPTION,
                    "properties": {
                        "code": {
                            "type": "integer",
                            "description": "The error code."
                        },
                        "message": {
                            "type": "string",
                            "description": "The error message."
                        }
                    },
                    "examples": [
                        {"code": 10, "message": "First test message"},
                        {"code": 15, "message": "Second est message"}
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)
        
        self.assertIn(ERROR, parser.object_schemas)
        self.assertEqual(parser.object_schemas[ERROR].description, ERROR_DESCRIPTION)
        self.assertEqual(parser.object_schemas[ERROR].title, 'Object title.')
        self.assertEqual(2, len(parser.object_schemas[ERROR].fields))
        self.assertIn('code', parser.object_schemas[ERROR].fields)
        self.assertIn('message', parser.object_schemas[ERROR].fields)

        CODE_REF_NAME = parser.object_schemas[ERROR].fields[0].type_reference
        self.assertIn(CODE_REF_NAME, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[CODE_REF_NAME].type, 'integer')
        self.assertEqual(parser.primitive_type_schemas[CODE_REF_NAME].description, 'The error code.')

        MESSAGE_REF_NAME = parser.object_schemas[ERROR].fields[1].type_reference
        self.assertIn(MESSAGE_REF_NAME, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[MESSAGE_REF_NAME].type, 'string')
        self.assertEqual(parser.primitive_type_schemas[MESSAGE_REF_NAME].description, 'The error message.')

        self.assertEqual(parser.object_schemas[ERROR].examples, 
                        [
                            {"code": 10, "message": "First test message"},
                            {"code": 15, "message": "Second est message"}
                        ])

        self.assertIsNone(parser.object_schemas[ERROR].required)
        self.assertIsNone(parser.object_schemas[ERROR].one_of_type_refs)
        self.assertIsNone(parser.object_schemas[ERROR].any_of_type_refs)
        self.assertIsNone(parser.object_schemas[ERROR].all_of_type_refs)
        self.assertIsNone(parser.object_schemas[ERROR].default)

    def test_object_with_required(self):
        EXECUTION_RESULT = "executionResult"
        EXECUTION_RESULT_DESCRIPTION = "The result of the execution."

        schema = {
            "definitions": {
                EXECUTION_RESULT: {
                    "type": "object",
                    "description": EXECUTION_RESULT_DESCRIPTION,
                    "properties": {
                        "success": {
                            "type": "boolean",
                            "description": "Success"
                        },
                        "error": {
                            "$ref": "#/definitions/error"
                        }
                    },
                    "required": [
                        "success"
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(EXECUTION_RESULT, parser.object_schemas)
        self.assertEqual(parser.object_schemas[EXECUTION_RESULT].description, EXECUTION_RESULT_DESCRIPTION)
        self.assertEqual(2, len(parser.object_schemas[EXECUTION_RESULT].fields))
        self.assertEqual(1, len(parser.object_schemas[EXECUTION_RESULT].required))
        self.assertIn('success', parser.object_schemas[EXECUTION_RESULT].fields)
        self.assertIn('error', parser.object_schemas[EXECUTION_RESULT].fields)
        self.assertIn('success', parser.object_schemas[EXECUTION_RESULT].required)

        SUCCESS_REF_NAME = parser.object_schemas[EXECUTION_RESULT].fields[0].type_reference
        self.assertIn(SUCCESS_REF_NAME, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[SUCCESS_REF_NAME].type, 'boolean')
        self.assertEqual(parser.primitive_type_schemas[SUCCESS_REF_NAME].description, 'Success')

        self.assertEqual(parser.object_schemas[EXECUTION_RESULT].fields[1].type_reference, 'error')

        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].one_of_type_refs)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].any_of_type_refs)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].all_of_type_refs)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].title)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].default)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].examples)


    def test_object_with_array(self):
        EXECUTION_RESULT = "executionResult"
        EXECUTION_RESULT_DESCRIPTION = "The result of the execution."

        schema = {
            "definitions": {
                EXECUTION_RESULT: {
                    "type": "object",
                    "description": EXECUTION_RESULT_DESCRIPTION,
                    "properties": {
                        "elements": {
                            "type": "array",
                            "description": "List of the elements.",
                            "items": {
                                "$ref": "#/definitions/elementId"
                            }
                        }
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)
        
        self.assertIn(EXECUTION_RESULT, parser.object_schemas)
        self.assertEqual(parser.object_schemas[EXECUTION_RESULT].description, EXECUTION_RESULT_DESCRIPTION)
        self.assertEqual(1, len(parser.object_schemas[EXECUTION_RESULT].fields))
        self.assertEqual(None, parser.object_schemas[EXECUTION_RESULT].required)
        self.assertIn('elements', parser.object_schemas[EXECUTION_RESULT].fields)

        ELEMENTS_REF_NAME = parser.object_schemas[EXECUTION_RESULT].fields[0].type_reference
        self.assertIn(ELEMENTS_REF_NAME, parser.list_schemas)
        self.assertEqual(parser.list_schemas[ELEMENTS_REF_NAME].description, "List of the elements.")
        self.assertEqual(parser.list_schemas[ELEMENTS_REF_NAME].itemtype, "elementId")

        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].required)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].one_of_type_refs)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].any_of_type_refs)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].all_of_type_refs)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].title)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].default)
        self.assertIsNone(parser.object_schemas[EXECUTION_RESULT].examples)

    def test_object_with_one_of(self):
        NAVIGATOR_TREE_ID = "NavigatorTreeId"
        NAVIGATOR_TREE_ID_DESCRIPTION = "The identifier of a navigator item tree."
        BOUNDING_BOX_2D_OR_ERROR = "BoundingBox2DOrError"
        BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION = "A 2D bounding box or error."
        
        schema = {
            "definitions": {
                NAVIGATOR_TREE_ID: {
                    "type": "object",
                    "description": NAVIGATOR_TREE_ID_DESCRIPTION,
                    "oneOf": [
                        {
                            "$ref": "#/definitions/PublisherSetId"
                        },
                        {
                            "$ref": "#/definitions/OtherNavigatorTreeId"
                        }
                    ]
                },
                BOUNDING_BOX_2D_OR_ERROR: {
                    "type": "object",
                    "description": BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION,
                    "oneOf": [
                        {
                            "properties": {
                                "boundingBox2D": {
                                    "$ref": "#/definitions/BoundingBox2D"
                                }
                            },
                            "required": [ "boundingBox2D" ]
                        },
                        {
                            "properties": {
                                "error": {
                                    "$ref": "#/definitions/Error"
                                }
                            },
                            "required": [ "error" ]
                        }
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(NAVIGATOR_TREE_ID, parser.object_schemas)
        self.assertIn(BOUNDING_BOX_2D_OR_ERROR, parser.object_schemas)

        self.assertEqual(parser.object_schemas[NAVIGATOR_TREE_ID].description, NAVIGATOR_TREE_ID_DESCRIPTION)
        self.assertEqual(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].description, BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].fields)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].fields)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].required)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].required)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].any_of_type_refs)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].all_of_type_refs)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs)
        
        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].title)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].title)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].default)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].default)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].examples)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].examples)


        self.assertIsNotNone(parser.object_schemas[NAVIGATOR_TREE_ID].one_of_type_refs)
        self.assertEqual(len(parser.object_schemas[NAVIGATOR_TREE_ID].one_of_type_refs), 2)
        self.assertEqual(['PublisherSetId', 'OtherNavigatorTreeId'], parser.object_schemas[NAVIGATOR_TREE_ID].one_of_type_refs)
        
        self.assertIsNotNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs)
        self.assertEqual(len(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs), 2)
        self.assertIn(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs[0], parser.object_schemas)
        self.assertIn(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs[1], parser.object_schemas)
        FIRST_OBJECT_TYPE = parser.object_schemas[parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs[0]]
        self.assertEqual(len(FIRST_OBJECT_TYPE.fields), 1)
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].name, 'boundingBox2D')
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].type_reference, 'BoundingBox2D')
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].type_reference, 'BoundingBox2D')
        self.assertEqual(len(FIRST_OBJECT_TYPE.required), 1)
        self.assertEqual(FIRST_OBJECT_TYPE.required[0], 'boundingBox2D')

        SECOND_OBJECT_TYPE = parser.object_schemas[parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs[1]]
        self.assertEqual(SECOND_OBJECT_TYPE.fields[0].name, 'error')
        self.assertEqual(SECOND_OBJECT_TYPE.fields[0].type_reference, 'Error')
        self.assertEqual(len(SECOND_OBJECT_TYPE.required), 1)
        self.assertEqual(SECOND_OBJECT_TYPE.required[0], 'error')


    def test_object_with_any_of(self):
        NAVIGATOR_TREE_ID = "NavigatorTreeId"
        NAVIGATOR_TREE_ID_DESCRIPTION = "The identifier of a navigator item tree."
        BOUNDING_BOX_2D_OR_ERROR = "BoundingBox2DOrError"
        BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION = "A 2D bounding box or error."

        schema = {
            "definitions": {
                NAVIGATOR_TREE_ID: {
                    "type": "object",
                    "description": NAVIGATOR_TREE_ID_DESCRIPTION,
                    "anyOf": [
                        {
                            "$ref": "#/definitions/PublisherSetId"
                        },
                        {
                            "$ref": "#/definitions/OtherNavigatorTreeId"
                        }
                    ]
                },
                BOUNDING_BOX_2D_OR_ERROR: {
                    "type": "object",
                    "description": BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION,
                    "anyOf": [
                        {
                            "properties": {
                                "boundingBox2D": {
                                    "$ref": "#/definitions/BoundingBox2D"
                                }
                            },
                            "required": [ "boundingBox2D" ]
                        },
                        {
                            "properties": {
                                "error": {
                                    "$ref": "#/definitions/Error"
                                }
                            },
                            "required": [ "error" ]
                        }
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(NAVIGATOR_TREE_ID, parser.object_schemas)
        self.assertIn(BOUNDING_BOX_2D_OR_ERROR, parser.object_schemas)

        self.assertEqual(parser.object_schemas[NAVIGATOR_TREE_ID].description, NAVIGATOR_TREE_ID_DESCRIPTION)
        self.assertEqual(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].description, BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].fields)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].fields)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].required)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].required)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].one_of_type_refs)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].all_of_type_refs)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs)
        
        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].title)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].title)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].default)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].default)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].examples)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].examples)
        
        self.assertIsNotNone(parser.object_schemas[NAVIGATOR_TREE_ID].any_of_type_refs)
        self.assertEqual(len(parser.object_schemas[NAVIGATOR_TREE_ID].any_of_type_refs), 2)
        self.assertEqual(['PublisherSetId', 'OtherNavigatorTreeId'], parser.object_schemas[NAVIGATOR_TREE_ID].any_of_type_refs)
        
        self.assertIsNotNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs)
        self.assertEqual(len(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs), 2)
        self.assertIn(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs[0], parser.object_schemas)
        self.assertIn(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs[1], parser.object_schemas)
        FIRST_OBJECT_TYPE = parser.object_schemas[parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs[0]]
        self.assertEqual(len(FIRST_OBJECT_TYPE.fields), 1)
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].name, 'boundingBox2D')
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].type_reference, 'BoundingBox2D')
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].type_reference, 'BoundingBox2D')
        self.assertEqual(len(FIRST_OBJECT_TYPE.required), 1)
        self.assertEqual(FIRST_OBJECT_TYPE.required[0], 'boundingBox2D')

        SECOND_OBJECT_TYPE = parser.object_schemas[parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs[1]]
        self.assertEqual(SECOND_OBJECT_TYPE.fields[0].name, 'error')
        self.assertEqual(SECOND_OBJECT_TYPE.fields[0].type_reference, 'Error')
        self.assertEqual(len(SECOND_OBJECT_TYPE.required), 1)
        self.assertEqual(SECOND_OBJECT_TYPE.required[0], 'error')
    
    def test_object_with_all_of(self):
        NAVIGATOR_TREE_ID = "NavigatorTreeId"
        NAVIGATOR_TREE_ID_DESCRIPTION = "The identifier of a navigator item tree."
        BOUNDING_BOX_2D_OR_ERROR = "BoundingBox2DOrError"
        BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION = "A 2D bounding box or error."

        schema = {
            "definitions": {
                NAVIGATOR_TREE_ID: {
                    "type": "object",
                    "description": NAVIGATOR_TREE_ID_DESCRIPTION,
                    "allOf": [
                        {
                            "$ref": "#/definitions/PublisherSetId"
                        },
                        {
                            "$ref": "#/definitions/OtherNavigatorTreeId"
                        }
                    ]
                },
                BOUNDING_BOX_2D_OR_ERROR: {
                    "type": "object",
                    "description": BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION,
                    "allOf": [
                        {
                            "properties": {
                                "boundingBox2D": {
                                    "$ref": "#/definitions/BoundingBox2D"
                                }
                            },
                            "required": [ "boundingBox2D" ]
                        },
                        {
                            "properties": {
                                "error": {
                                    "$ref": "#/definitions/Error"
                                }
                            },
                            "required": [ "error" ]
                        }
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(NAVIGATOR_TREE_ID, parser.object_schemas)
        self.assertIn(BOUNDING_BOX_2D_OR_ERROR, parser.object_schemas)

        self.assertEqual(parser.object_schemas[NAVIGATOR_TREE_ID].description, NAVIGATOR_TREE_ID_DESCRIPTION)
        self.assertEqual(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].description, BOUNDING_BOX_2D_OR_ERROR_DESCRIPTION)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].fields)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].fields)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].required)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].required)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].one_of_type_refs)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].one_of_type_refs)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].any_of_type_refs)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].any_of_type_refs)
        
        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].title)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].title)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].default)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].default)

        self.assertIsNone(parser.object_schemas[NAVIGATOR_TREE_ID].examples)
        self.assertIsNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].examples)
        
        self.assertIsNotNone(parser.object_schemas[NAVIGATOR_TREE_ID].all_of_type_refs)
        self.assertEqual(len(parser.object_schemas[NAVIGATOR_TREE_ID].all_of_type_refs), 2)
        self.assertEqual(['PublisherSetId', 'OtherNavigatorTreeId'], parser.object_schemas[NAVIGATOR_TREE_ID].all_of_type_refs)
        
        self.assertIsNotNone(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs)
        self.assertEqual(len(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs), 2)
        self.assertIn(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs[0], parser.object_schemas)
        self.assertIn(parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs[1], parser.object_schemas)
        FIRST_OBJECT_TYPE = parser.object_schemas[parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs[0]]
        self.assertEqual(len(FIRST_OBJECT_TYPE.fields), 1)
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].name, 'boundingBox2D')
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].type_reference, 'BoundingBox2D')
        self.assertEqual(FIRST_OBJECT_TYPE.fields[0].type_reference, 'BoundingBox2D')
        self.assertEqual(len(FIRST_OBJECT_TYPE.required), 1)
        self.assertEqual(FIRST_OBJECT_TYPE.required[0], 'boundingBox2D')

        SECOND_OBJECT_TYPE = parser.object_schemas[parser.object_schemas[BOUNDING_BOX_2D_OR_ERROR].all_of_type_refs[1]]
        self.assertEqual(SECOND_OBJECT_TYPE.fields[0].name, 'error')
        self.assertEqual(SECOND_OBJECT_TYPE.fields[0].type_reference, 'Error')
        self.assertEqual(len(SECOND_OBJECT_TYPE.required), 1)
        self.assertEqual(SECOND_OBJECT_TYPE.required[0], 'error')
    

class TestArrays(unittest.TestCase):
    def test_bool_array(self):
        BOOL_ARRAY = "boolArray"
        BOOL_ARRAY_DESCRIPTION = 'An array of booleans.'

        schema = {
            "definitions": {
                BOOL_ARRAY: {
                    "title": "Test bool array title.",
                    "type": "array",
                    "description": BOOL_ARRAY_DESCRIPTION,
                    "items": {
                        "type": "boolean",
                    },
                    "examples": [
                        [True, True, False],
                        [False, False, False]
                    ]
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(BOOL_ARRAY, parser.list_schemas)
        self.assertEqual(parser.list_schemas[BOOL_ARRAY].description, BOOL_ARRAY_DESCRIPTION)
        self.assertEqual(parser.list_schemas[BOOL_ARRAY].title, 'Test bool array title.')
        self.assertEqual(parser.list_schemas[BOOL_ARRAY].examples, 
                        [
                            [True, True, False],
                            [False, False, False]
                        ])
        self.assertIsNone(parser.list_schemas[BOOL_ARRAY].default)
        
        BOOL_REF_NAME = parser.list_schemas[BOOL_ARRAY].itemtype
        self.assertEqual(parser.primitive_type_schemas[BOOL_REF_NAME].type, 'boolean')

    def test_number_array(self):
        NUMBER_ARRAY = "numberArray"

        schema = {
            "definitions": {
                NUMBER_ARRAY: {
                    "type": "array",
                    "items": {
                        "type": "number",
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(NUMBER_ARRAY, parser.list_schemas)
        self.assertFalse(parser.list_schemas[NUMBER_ARRAY].description)
        self.assertIsNone(parser.list_schemas[NUMBER_ARRAY].title)
        self.assertIsNone(parser.list_schemas[NUMBER_ARRAY].default)
        self.assertIsNone(parser.list_schemas[NUMBER_ARRAY].examples)
        
        INTEGER_REF_NAME = parser.list_schemas[NUMBER_ARRAY].itemtype
        self.assertEqual(parser.primitive_type_schemas[INTEGER_REF_NAME].type, 'number')
        
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].maximum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].exclusive_maximum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].minimum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].exclusive_minimum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].values)

    def test_int_array(self):
        INTEGER_ARRAY = "integerArray"
        INTEGER_ARRAY_DESCRIPTION = "An array of integers."
        INTEGER_ARRAY_ITEMS_DESCRIPTION = "An integer of the array."

        schema = {
            "definitions": {
                INTEGER_ARRAY: {
                    "type": "array",
                    "description": INTEGER_ARRAY_DESCRIPTION,
                    "items": {
                        "type": "integer",
                        "description": INTEGER_ARRAY_ITEMS_DESCRIPTION
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(INTEGER_ARRAY, parser.list_schemas)
        self.assertEqual(parser.list_schemas[INTEGER_ARRAY].description, INTEGER_ARRAY_DESCRIPTION)
        self.assertIsNone(parser.list_schemas[INTEGER_ARRAY].title)
        self.assertIsNone(parser.list_schemas[INTEGER_ARRAY].default)
        self.assertIsNone(parser.list_schemas[INTEGER_ARRAY].examples)
        
        INTEGER_REF_NAME = parser.list_schemas[INTEGER_ARRAY].itemtype
        self.assertEqual(parser.primitive_type_schemas[INTEGER_REF_NAME].type, 'integer')
        self.assertEqual(parser.primitive_type_schemas[INTEGER_REF_NAME].description, INTEGER_ARRAY_ITEMS_DESCRIPTION)
        
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].maximum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].exclusive_maximum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].minimum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].exclusive_minimum)
        self.assertIsNone(parser.primitive_type_schemas[INTEGER_REF_NAME].values)

    def test_guid_array(self):
        GUID = "guid"
        GUID_PATTERN = "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"
        GUID_DESCRIPTION = "A Globally Unique Identifier (or Universally Unique Identifier) in its string representation as defined in RFC 4122."
        GUID_ARRAY = "guidArray"
        GUID_ARRAY_DESCRIPTION = "An array of guids."

        schema = {
            "definitions": {
                GUID: {
                    "type": "string",
                    "description": GUID_DESCRIPTION,
                    "format": "uuid",
                    "pattern": GUID_PATTERN
                },
                GUID_ARRAY: {
                    "type": "array",
                    "description": GUID_ARRAY_DESCRIPTION,
                    "items": {
                        "$ref": "#/definitions/guid"
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(GUID_ARRAY, parser.list_schemas)
        self.assertEqual(parser.list_schemas[GUID_ARRAY].description, GUID_ARRAY_DESCRIPTION)
        self.assertIsNone(parser.list_schemas[GUID_ARRAY].title)
        self.assertIsNone(parser.list_schemas[GUID_ARRAY].default)
        self.assertIsNone(parser.list_schemas[GUID_ARRAY].examples)
        
        self.assertEqual(parser.list_schemas[GUID_ARRAY].itemtype, GUID)
        GUID_REF_NAME = parser.list_schemas[GUID_ARRAY].itemtype
        self.assertEqual(parser.primitive_type_schemas[GUID_REF_NAME].type, 'UUID')
        self.assertEqual(parser.primitive_type_schemas[GUID_REF_NAME].description, GUID_DESCRIPTION)

        self.assertIsNotNone(parser.primitive_type_schemas[GUID_REF_NAME].pattern)
        self.assertEqual(parser.primitive_type_schemas[GUID_REF_NAME].pattern, GUID_PATTERN)

        self.assertIsNone(parser.primitive_type_schemas[GUID_REF_NAME].min_length)
        self.assertIsNone(parser.primitive_type_schemas[GUID_REF_NAME].max_length)
        self.assertIsNone(parser.primitive_type_schemas[GUID_REF_NAME].values)

    def test_string_array(self):
        STRING_ARRAY = "stringArray"
        STRING_ARRAY_DESCRIPTION = "An array of strings."
        STRING_ARRAY_ITEMS_DESCRIPTION = "A string of the array."

        schema = {
            "definitions": {
                STRING_ARRAY: {
                    "type": "array",
                    "description": STRING_ARRAY_DESCRIPTION,
                    "items": {
                        "type": "string",
                        "description": STRING_ARRAY_ITEMS_DESCRIPTION
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(STRING_ARRAY, parser.list_schemas)
        self.assertEqual(parser.list_schemas[STRING_ARRAY].description, STRING_ARRAY_DESCRIPTION)
        self.assertIsNone(parser.list_schemas[STRING_ARRAY].title)
        self.assertIsNone(parser.list_schemas[STRING_ARRAY].default)
        self.assertIsNone(parser.list_schemas[STRING_ARRAY].examples)
        
        STR_REF_NAME = parser.list_schemas[STRING_ARRAY].itemtype
        self.assertEqual(parser.primitive_type_schemas[STR_REF_NAME].type, 'string')
        self.assertEqual(parser.primitive_type_schemas[STR_REF_NAME].description, STRING_ARRAY_ITEMS_DESCRIPTION)
    
    def test_predefined_object_array(self):
        ERROR = "error"
        ERROR_DESCRIPTION = "Error details."
        OBJECT_ARRAY = "objectArray"
        OBJECT_ARRAY_DESCRIPTION = "An array of errors."

        schema = {
            "definitions": {
                ERROR: {
                    "type": "object",
                    "description": ERROR_DESCRIPTION,
                    "properties": {
                        "code": {
                            "type": "integer",
                            "description": "The error code."
                        },
                        "message": {
                            "type": "string",
                            "description": "The error message."
                        }
                    },
                    "required": [
                        "code",
                        "message"
                    ]
                },
                OBJECT_ARRAY: {
                    "type": "array",
                    "description": OBJECT_ARRAY_DESCRIPTION,
                    "items": {
                        "$ref": "#/definitions/error"
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(OBJECT_ARRAY, parser.list_schemas)
        self.assertEqual(parser.list_schemas[OBJECT_ARRAY].description, OBJECT_ARRAY_DESCRIPTION)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].title)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].default)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].examples)
        
        self.assertIn(ERROR, parser.object_schemas)
        self.assertEqual(parser.object_schemas[ERROR].description, ERROR_DESCRIPTION)
        self.assertEqual(2, len(parser.object_schemas[ERROR].fields))
        self.assertEqual(2, len(parser.object_schemas[ERROR].required))
        self.assertIn('code', parser.object_schemas[ERROR].fields)
        self.assertIn('message', parser.object_schemas[ERROR].fields)
        self.assertIn('code', parser.object_schemas[ERROR].required)
        self.assertIn('message', parser.object_schemas[ERROR].required)

        CODE_REF_NAME = parser.object_schemas[ERROR].fields[0].type_reference
        self.assertIn(CODE_REF_NAME, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[CODE_REF_NAME].type, 'integer')
        self.assertEqual(parser.primitive_type_schemas[CODE_REF_NAME].description, 'The error code.')

        MESSAGE_REF_NAME = parser.object_schemas[ERROR].fields[1].type_reference
        self.assertIn(MESSAGE_REF_NAME, parser.primitive_type_schemas)
        self.assertEqual(parser.primitive_type_schemas[MESSAGE_REF_NAME].type, 'string')
        self.assertEqual(parser.primitive_type_schemas[MESSAGE_REF_NAME].description, 'The error message.')
        
    def test_undefined_object_array(self):
        OBJECT_ARRAY = "objectArray"
        OBJECT_ARRAY_DESCRIPTION = "An array of undefined objects."

        schema = {
            "definitions": {
                OBJECT_ARRAY: {
                    "type": "array",
                    "description": OBJECT_ARRAY_DESCRIPTION,
                    "items": {
                        "type": "object",
                        "properties": {
                            "navigatorItemId": {
                                "$ref": "#/definitions/NavigatorItemId"
                            }
                        },
                        "required": [
                            "navigatorItemId"
                        ]
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)
        
        self.assertIn(OBJECT_ARRAY, parser.list_schemas)
        self.assertEqual(parser.list_schemas[OBJECT_ARRAY].description, OBJECT_ARRAY_DESCRIPTION)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].title)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].default)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].examples)

        ITEM_TYPE_REF: str = parser.list_schemas[OBJECT_ARRAY].itemtype
        self.assertRegex(ITEM_TYPE_REF, '^Object_[0-9a-fA-F]{8}_[0-9a-fA-F]{4}_[0-9a-fA-F]{4}_[0-9a-fA-F]{4}_[0-9a-fA-F]{12}$')
        self.assertEqual(len(parser.object_schemas[ITEM_TYPE_REF].fields), 1)
        self.assertEqual(parser.object_schemas[ITEM_TYPE_REF].fields[0].name, 'navigatorItemId')
        self.assertEqual(parser.object_schemas[ITEM_TYPE_REF].fields[0].type_reference, 'NavigatorItemId')
        self.assertEqual(len(parser.object_schemas[ITEM_TYPE_REF].required), 1)
        self.assertIn('navigatorItemId', parser.object_schemas[ITEM_TYPE_REF].required)

        self.assertEqual(parser.object_schemas[ITEM_TYPE_REF].description, '')
        self.assertIsNone(parser.object_schemas[ITEM_TYPE_REF].one_of_type_refs)
        self.assertIsNone(parser.object_schemas[ITEM_TYPE_REF].any_of_type_refs)
        self.assertIsNone(parser.object_schemas[ITEM_TYPE_REF].all_of_type_refs)

        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].min_items)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].max_items)
        self.assertIsNone(parser.list_schemas[OBJECT_ARRAY].unique_items)

    def test_full_array_schema(self):
        GUID_ARRAY = "guidArray"
        GUID_ARRAY_DESCRIPTION = "An array of guids."

        schema = {
            "definitions": {
                GUID_ARRAY: {
                    "title": "Test title.",
                    "type": "array",
                    "description": GUID_ARRAY_DESCRIPTION,
                    "items": {
                        "$ref": "#/definitions/guid"
                    },
                    "minItems": 5,
                    "maxItems": 10,
                    "uniqueItems": True
                }
            }
        }
        parser = SchemaParser()
        parser.parse_types(schema)

        self.assertIn(GUID_ARRAY, parser.list_schemas)
        self.assertEqual(parser.list_schemas[GUID_ARRAY].description, GUID_ARRAY_DESCRIPTION)
        self.assertEqual(parser.list_schemas[GUID_ARRAY].itemtype, 'guid')
        self.assertEqual(parser.list_schemas[GUID_ARRAY].title, 'Test title.')
        self.assertIsNone(parser.list_schemas[GUID_ARRAY].examples)
        self.assertIsNone(parser.list_schemas[GUID_ARRAY].default)

        self.assertEqual(parser.list_schemas[GUID_ARRAY].min_items, 5)
        self.assertEqual(parser.list_schemas[GUID_ARRAY].max_items, 10)
        self.assertEqual(parser.list_schemas[GUID_ARRAY].unique_items, True)

class TestCommands(unittest.TestCase):
    def test_command(self):
        COMMAND_NAME = "CreateViewMapFolder"
        COMMAND_DESCRIPTION = "Create a view folder item at a given position in the navigator tree."

        schema= {
            "$schema": "http://json-schema.org/draft-04/schema",
            "description": COMMAND_DESCRIPTION,
            "definitions": {
                "command_parameters": {
                    "type": "object",
                    "properties": {
                        "folderParameters": {
                            "$ref": "APITypes.json#/definitions/FolderParameters"
                        },
                        "parentFolder": {
                            "$ref": "APITypes.json#/definitions/NavigatorItemId"
                        },
                        "previousSiblingFolder": {
                            "$ref": "APITypes.json#/definitions/NavigatorItemId"
                        }
                    },
                    "required": [
                        "folderParameters"
                    ]
                },
                "response_parameters": {
                    "type": "object",
                    "properties": {
                    }
                },
                "examples": {
                    "command_example": {
                        "command": "API.CreateFolderInViewTree",
                        "parameters": {
                            "folderParameters": {
                                "folderName": "Test Folder"
                            },
                            "parentFolder": {
                                "guid": "8A3BBC57-86D6-407E-8B3C-3A862E32DF3A"
                            },
                            "previousSiblingFolder": {
                                "guid": "00000000-4315-4F9E-84F5-275194D29226"
                            }
                        }
                    },
                    "response_example": {
                        "succeeded": True,
                        "result": {
                            "createdFolderId": {
                                "guid": "EDB6197B-0754-4741-8BA6-712865FD0F76"
                            }
                        },
                        "executionDuration": 0.242254153
                    }
                }
            }
        }
        parser = SchemaParser()
        parser.parse_command(COMMAND_NAME, schema)

        self.assertIn(COMMAND_NAME, parser.command_schemas)
        COMMAND = parser.command_schemas[COMMAND_NAME]

        self.assertEqual(COMMAND.description, COMMAND_DESCRIPTION)
        self.assertIn(COMMAND.parameters_schema_name, parser.object_schemas)
        self.assertEqual(COMMAND.parameters_schema_name, f'{COMMAND_NAME}_parameters')
        self.assertIn(COMMAND.response_schema_name, parser.object_schemas)
        self.assertEqual(COMMAND.response_schema_name, f'{COMMAND_NAME}_response')

        PARAMETER_OBJECT = parser.object_schemas[COMMAND.parameters_schema_name]
        self.assertEqual(len(PARAMETER_OBJECT.fields), 3)
        self.assertEqual(PARAMETER_OBJECT.fields[0].name, 'folderParameters')
        self.assertEqual(PARAMETER_OBJECT.fields[0].type_reference, 'FolderParameters')

        self.assertEqual(PARAMETER_OBJECT.fields[1].name, 'parentFolder')
        self.assertEqual(PARAMETER_OBJECT.fields[1].type_reference, 'NavigatorItemId')

        self.assertEqual(PARAMETER_OBJECT.fields[2].name, 'previousSiblingFolder')
        self.assertEqual(PARAMETER_OBJECT.fields[2].type_reference, 'NavigatorItemId')

        self.assertIn('folderParameters', PARAMETER_OBJECT.required)

        self.assertEqual(PARAMETER_OBJECT.description, '')
        self.assertIsNone(PARAMETER_OBJECT.one_of_type_refs)
        self.assertIsNone(PARAMETER_OBJECT.any_of_type_refs)
        self.assertIsNone(PARAMETER_OBJECT.all_of_type_refs)

        RESPONSE_OBJECT = parser.object_schemas[COMMAND.response_schema_name]
        self.assertEqual(RESPONSE_OBJECT.description, '')
        self.assertIsNone(RESPONSE_OBJECT.fields)
        self.assertIsNone(RESPONSE_OBJECT.required)
        self.assertIsNone(RESPONSE_OBJECT.one_of_type_refs)
        self.assertIsNone(RESPONSE_OBJECT.any_of_type_refs)
        self.assertIsNone(RESPONSE_OBJECT.all_of_type_refs)
        
        self.assertEqual(COMMAND.examples,
                        {
                            "command_example": {
                                "command": "API.CreateFolderInViewTree",
                                "parameters": {
                                    "folderParameters": {
                                        "folderName": "Test Folder"
                                    },
                                    "parentFolder": {
                                        "guid": "8A3BBC57-86D6-407E-8B3C-3A862E32DF3A"
                                    },
                                    "previousSiblingFolder": {
                                        "guid": "00000000-4315-4F9E-84F5-275194D29226"
                                    }
                                }
                            },
                            "response_example": {
                                "succeeded": True,
                                "result": {
                                    "createdFolderId": {
                                        "guid": "EDB6197B-0754-4741-8BA6-712865FD0F76"
                                    }
                                },
                                "executionDuration": 0.242254153
                            }
                        })




"""GRAPHISOFT"""

from typing import Dict, Any, Tuple
from urllib.request import Request, urlopen
import json


class UnsucceededCommandCall(Exception):
    pass


def post_command(req: Request, jsonStr: str) -> Dict[str, Any]:
    response = urlopen(req, jsonStr.encode("UTF-8"))
    result = response.read()
    return json.loads(result)


class BasicCommands:
    """Collection of the ARCHICAD JSON interface commands
    """
    def __init__(self, req: Request):
        assert req is not None
        self.__req = req

    def IsAlive(self) -> bool:
        """Checks if the ARCHICAD connection is alive.

        Returns:
            bool: Returns true if the connection is alive.

        """

        result = post_command(self.__req, json.dumps({"command": "API.IsAlive"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return result["result"]["isAlive"]

    def GetProductInfo(self) -> Tuple[int, int, str]:
        """Accesses the version information from the running ARCHICAD.

        Returns:
            int: The version of the running ARCHICAD.
            int: The build number of the running ARCHICAD.
            str: The language code of the running ARCHICAD.

        """

        result = post_command(self.__req, json.dumps({"command": "API.GetProductInfo"}))
        if not result["succeeded"]:
            raise UnsucceededCommandCall(result)
        return result["result"]["version"], result["result"]["buildNumber"], result["result"]["languageCode"]



"""Typing framework for the ARCHICAD JSON interface
"""

from typing import List, Dict, ClassVar, Any, Callable, Union, Optional, Set, Tuple
from enum import Enum, EnumMeta
from uuid import UUID
import abc

_GENERIC_TYPE_ = type(List)

def is_class(typ) -> bool:
    """Returns True if the parameter is a class type
    """
    return isinstance(typ, type) and typ is not type(None)


def is_generic_list(typ) -> bool:
    """Returns True if the parameter is a generic list
    """
    return type(typ) is _GENERIC_TYPE_ and typ.__origin__ is list


def is_union(typ) -> bool:
    return type(typ) is _GENERIC_TYPE_ and typ.__origin__ is Union


def none_constructor(*args, **kwargs) -> None:
    if not args and not kwargs:
        return None
    if len(args) == 1 and len(kwargs) == 0 and args[0] is None:
        return None
    raise TypeError('"None" can\'t be initialized with parameters')


def acbasetype_constructor_wrapper(acbasetyp: type):
    typ = acbasetyp
    def constructor(*args, **kwargs):
        if not args:
            return typ(**dict({k:v for k,v in kwargs.items() if v is not None}))
        if args and kwargs:
            # Ha valaki a sok None-t ki akarja hagyni, előfordulhat ez az eset is.
            # pl: ehelyett: (navigatorItemId, None, myNewId)  
            #     ez:       (navigatorItemId, newId=myNewId)
            return typ(*tuple(arg for arg in args if arg is not None), **dict({k:v for k,v in kwargs.items() if v is not None}))
        if isinstance(args[0], dict):
            return typ(**args[0])
        if isinstance(args[0], typ):
            return args[0]
        if issubclass(typ, _ACUnionType) and type(args[0]) in typ.constructor.optional_types:
            return args[0]
        if not kwargs:
            # Jelenlegi generált kódban tartva van a sorrend,
            # tehát ha a None-okat kiszedjük, akkor probléma mentesen működik.
            return typ(*tuple(arg for arg in args if arg is not None))
        raise TypeError(f'{typ} can\'t be initialized with theese parameters.')
    return constructor


def _get_constructor(typ) -> Callable:
    """Returns the corresponding constructor callable to the type.
    """
    if is_generic_list(typ):
        return _ListBuilder(typ.__args__[0])
    if is_union(typ):
        return _ConstructUnion(typ)
    if is_class(typ):
        if issubclass(typ, _ACBaseType):
            return acbasetype_constructor_wrapper(typ)
        else:
            return typ
    if typ is type(None):
        return none_constructor
    raise NotImplementedError()


class _ListBuilder:
    """A constructor object for the type-annotated generic lists
    """
    def __init__(self, item_type):
        self.constructor = _get_constructor(item_type)

    def __call__(self, items: list):
        assert isinstance(items, list)
        return [self.constructor(item) for item in items]


class _ConstructUnion:
    __slots__ = ('optional_types',)

    def __init__(self, union_type: Union[Any]):
        if is_union(union_type):
            self.optional_types: tuple = union_type.__args__
        elif is_class(union_type):
            self.optional_types = (union_type,)

    def __call__(self, *args, **kwargs):
        valid_results = []
        for one_type in self.optional_types:
            try:
                valid_results.append(_get_constructor(one_type)(*args, **kwargs))
            except Exception as e:
                pass
        if not valid_results:
            raise TypeError("This initialization is not applicable for this union type")
        assert len(valid_results) == 1
        return valid_results[0]


class _ClassInfo:    
    def __init__(self):
        self.fields: Dict[str, Any] = {}
        self.value_validators: Dict[str, Tuple[Callable[[Any], bool], ...]] = {}
        self.instance_validators: Tuple[Callable[[Any], bool], ...] = ()

    def __contains__(self, name: str):
        return name in self.fields

    def add_field(self, field_name: str, field_type: Any, *validators: Callable[[Any], bool]):
        assert field_name not in self
        assert is_class(field_type) or is_generic_list(field_type) or is_union(field_type)
        self.fields[field_name] = field_type
        if validators:
            self.value_validators[field_name] = validators
        
    def is_valid_value(self, name: str, value: Any) -> bool:
        if value is None:
            return True
        return all([validator(value) for validator in self.value_validators.get(name, ())])
    
    def add_instance_validators(self, *validators: Callable[[Any], bool]):
        self.instance_validators = validators
    
    @staticmethod
    def is_valid_instance(obj):
        return all(validator(obj) for validator in obj.get_classinfo().instance_validators)


class _ACBaseType(abc.ABC):
    _class_info: ClassVar[_ClassInfo]

    __slots__ = ()

    @classmethod
    def get_classinfo(cls) -> _ClassInfo:
        """Returns the ClassInfo object of the type. If not exists, it will be created.
        """
        if '_class_info' not in cls.__dict__:
            cls._class_info = _ClassInfo()
        return cls._class_info

    # member functions

    def __setattr__(self, name, value):
        attr_type = self.get_classinfo().fields[name]

        type_match = isinstance(attr_type, type) and isinstance(value, attr_type)
        is_union_instance = (
            (is_union(attr_type) and type(value) in attr_type.__args__) or
            (is_class(attr_type) and issubclass(attr_type, _ACUnionType) and type(value) in attr_type.constructor.optional_types)
        )
        is_actype_with_dict = is_class(attr_type) and issubclass(attr_type, _ACBaseType) and isinstance(value, dict)
        is_list = is_generic_list(attr_type) and isinstance(value, list)
        is_uuid = attr_type is UUID and isinstance(value, str)
        int_to_float = attr_type == float and type(value) == int

        value_to_set = None
        if type_match or is_union_instance:
            value_to_set = value
        elif is_list or is_uuid or is_union(attr_type) or is_actype_with_dict or int_to_float:
            value_to_set = _get_constructor(attr_type)(value)
        else:
            raise TypeError(f'{name} cannot be initialized with {type(value)}')

        if self.get_classinfo().is_valid_value(name, value_to_set):
            super().__setattr__(name, value_to_set)
        else:
            raise ValueError(f"{value} is not allowed for {name}")


    def to_dict(self) -> Dict[str, Any]:
        """Returns the dict representation of the object.
        """
        def deserialize(value):
            if isinstance(value, _ACBaseType):
                return value.to_dict()
            if isinstance(value, UUID):
                return str(value).upper()
            if isinstance(value, list):
                return [deserialize(item) for item in value]
            return value

        return {fieldname: deserialize(self.__getattribute__(fieldname))
                    for fieldname in self.__slots__
                        if self.__getattribute__(fieldname) is not None}

    def __repr__(self):
        return type(self).__name__ + ' ' + str(self.to_dict())

class _ACUnionType(_ACBaseType):
    constructor: _ConstructUnion



"""GRAPHISOFT
"""
from uuid import UUID
from typing import Union, Optional, List

from archicad.acbasetype import _ACBaseType, _ACUnionType, _ConstructUnion
from archicad.validators import value_set, matches, min_length, max_length, multiple_of, minimum, maximum, listitem_validator, min_items, max_items, unique_items


class ClassificationSystemId(_ACBaseType):
    """ The identifier of a classification system.

    Attributes:
        guid (:obj:`UUID`): A Globally Unique Identifier (or Universally Unique Identifier) in its string representation as defined in RFC 4122.

    """
    __slots__ = ("guid", )

    def __init__(self, guid: UUID):
        self.guid: UUID = guid

ClassificationSystemId.get_classinfo().add_field('guid', UUID)


class ClassificationSystemIdArrayItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        classificationSystemId (:obj:`ClassificationSystemId`): The identifier of a classification system.

    """
    __slots__ = ("classificationSystemId", )

    def __init__(self, classificationSystemId: ClassificationSystemId):
        self.classificationSystemId: ClassificationSystemId = classificationSystemId

ClassificationSystemIdArrayItem.get_classinfo().add_field('classificationSystemId', ClassificationSystemId)


class ClassificationItemId(_ACBaseType):
    """ The identifier of a classification item.

    Attributes:
        guid (:obj:`UUID`): A Globally Unique Identifier (or Universally Unique Identifier) in its string representation as defined in RFC 4122.

    """
    __slots__ = ("guid", )

    def __init__(self, guid: UUID):
        self.guid: UUID = guid

ClassificationItemId.get_classinfo().add_field('guid', UUID)


class ClassificationItemIdArrayItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        classificationItemId (:obj:`ClassificationItemId`): The identifier of a classification item.

    """
    __slots__ = ("classificationItemId", )

    def __init__(self, classificationItemId: ClassificationItemId):
        self.classificationItemId: ClassificationItemId = classificationItemId

ClassificationItemIdArrayItem.get_classinfo().add_field('classificationItemId', ClassificationItemId)


class ClassificationId(_ACBaseType):
    """ The element classification identifier.

    Attributes:
        classificationSystemId (:obj:`ClassificationSystemId`): The identifier of a classification system.
        classificationItemId (:obj:`ClassificationItemId`, optional): The identifier of a classification item.

    """
    __slots__ = ("classificationSystemId", "classificationItemId", )

    def __init__(self, classificationSystemId: ClassificationSystemId, classificationItemId: Optional[ClassificationItemId] = None):
        self.classificationSystemId: ClassificationSystemId = classificationSystemId
        self.classificationItemId: Optional[ClassificationItemId] = classificationItemId

ClassificationId.get_classinfo().add_field('classificationSystemId', ClassificationSystemId)
ClassificationId.get_classinfo().add_field('classificationItemId', Optional[ClassificationItemId])


class ClassificationItemDetails(_ACBaseType):
    """ Details of a classification item.

    Attributes:
        id (:obj:`str`): The user specified unique identifier of the classification item.
        name (:obj:`str`): The display name of the classification item.
        description (:obj:`str`): The description of the classification item.

    """
    __slots__ = ("id", "name", "description", )

    def __init__(self, id: str, name: str, description: str):
        self.id: str = id
        self.name: str = name
        self.description: str = description

ClassificationItemDetails.get_classinfo().add_field('id', str)
ClassificationItemDetails.get_classinfo().add_field('name', str)
ClassificationItemDetails.get_classinfo().add_field('description', str)


class ClassificationSystem(_ACBaseType):
    """ Details of a classification system.

    Attributes:
        classificationSystemId (:obj:`ClassificationSystemId`): The identifier of a classification system.
        name (:obj:`str`): The display name of the classification system.
        description (:obj:`str`): The description of the classification system.
        source (:obj:`str`): The source of the classification system (e.g. URL to a classification system standard).
        version (:obj:`str`): The version string of the classification system.
        date (:obj:`str`): A date in its string representation as defined in ISO 8601: YYYY-MM-DD.

    """
    __slots__ = ("classificationSystemId", "name", "description", "source", "version", "date", )

    def __init__(self, classificationSystemId: ClassificationSystemId, name: str, description: str, source: str, version: str, date: str):
        self.classificationSystemId: ClassificationSystemId = classificationSystemId
        self.name: str = name
        self.description: str = description
        self.source: str = source
        self.version: str = version
        self.date: str = date

ClassificationSystem.get_classinfo().add_field('classificationSystemId', ClassificationSystemId)
ClassificationSystem.get_classinfo().add_field('name', str)
ClassificationSystem.get_classinfo().add_field('description', str)
ClassificationSystem.get_classinfo().add_field('source', str)
ClassificationSystem.get_classinfo().add_field('version', str)
ClassificationSystem.get_classinfo().add_field('date', str, matches(r"^[0-9]{4}-[0-9]{2}-[0-9]{2}$"))


class UserDefinedPropertyUserId(_ACBaseType):
    """ An object which uniquely identifies a User-Defined Property by its name in human-readable form.

    Attributes:
        localizedName (:obj:`list` of :obj:`str`): List of the localized name parts: first element is the Group Name, second element is the Property Name of the Property.
        type (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("localizedName", "type", )

    def __init__(self, localizedName: List[str], type: str = "UserDefined"):
        self.localizedName: List[str] = localizedName
        self.type: str = type

UserDefinedPropertyUserId.get_classinfo().add_field('localizedName', List[str], min_items(2), max_items(2))
UserDefinedPropertyUserId.get_classinfo().add_field('type', str, value_set(['UserDefined']))


class BuiltInPropertyUserId(_ACBaseType):
    """ An object which uniquely identifies a Built-In Property by its name in a human-readable form.

    Attributes:
        nonLocalizedName (:obj:`str`): Non-localized name of the Built-In Property.
        type (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("nonLocalizedName", "type", )

    def __init__(self, nonLocalizedName: str, type: str = "BuiltIn"):
        self.nonLocalizedName: str = nonLocalizedName
        self.type: str = type

BuiltInPropertyUserId.get_classinfo().add_field('nonLocalizedName', str)
BuiltInPropertyUserId.get_classinfo().add_field('type', str, value_set(['BuiltIn']))


class PropertyUserId(_ACUnionType):
    """ An object which uniquely identifies a Property by its name in human-readable form. May represent a User-Defined or a Built-In Property.

    Attributes:
        type (:obj:`str`): None
        localizedName (:obj:`list` of :obj:`str`, optional): List of the localized name parts: first element is the Group Name, second element is the Property Name of the Property.
        nonLocalizedName (:obj:`str`, optional): Non-localized name of the Built-In Property.

    """
    __slots__ = ("type", "localizedName", "nonLocalizedName", )

    constructor  = _ConstructUnion(Union[UserDefinedPropertyUserId, BuiltInPropertyUserId])

    def __new__(cls, type: str, localizedName: Optional[List[str]] = None, nonLocalizedName: Optional[str] = None):
        return cls.constructor(type=type, localizedName=localizedName, nonLocalizedName=nonLocalizedName)

    def __init__(self, type: str, localizedName: Optional[List[str]] = None, nonLocalizedName: Optional[str] = None):
        self.type: str = type
        self.localizedName: Optional[List[str]] = localizedName
        self.nonLocalizedName: Optional[str] = nonLocalizedName


class PropertyId(_ACBaseType):
    """ The identifier of a property.

    Attributes:
        guid (:obj:`UUID`): A Globally Unique Identifier (or Universally Unique Identifier) in its string representation as defined in RFC 4122.

    """
    __slots__ = ("guid", )

    def __init__(self, guid: UUID):
        self.guid: UUID = guid

PropertyId.get_classinfo().add_field('guid', UUID)


class PropertyIdArrayItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        propertyId (:obj:`PropertyId`): The identifier of a property.

    """
    __slots__ = ("propertyId", )

    def __init__(self, propertyId: PropertyId):
        self.propertyId: PropertyId = propertyId

PropertyIdArrayItem.get_classinfo().add_field('propertyId', PropertyId)


class PropertyGroup(_ACBaseType):
    """ A property group.

    Attributes:
        name (:obj:`str`): The property group name.

    """
    __slots__ = ("name", )

    def __init__(self, name: str):
        self.name: str = name

PropertyGroup.get_classinfo().add_field('name', str)


class NormalNumberPropertyValue(_ACBaseType):
    """ A number property value containing a valid numeric value.

    Attributes:
        value (:obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: float, type: str = "number", status: str = "normal"):
        self.value: float = value
        self.type: str = type
        self.status: str = status

NormalNumberPropertyValue.get_classinfo().add_field('value', float)
NormalNumberPropertyValue.get_classinfo().add_field('type', str, value_set(['number']))
NormalNumberPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalIntegerPropertyValue(_ACBaseType):
    """ An integer property value containing a valid integer number.

    Attributes:
        value (:obj:`int`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: int, type: str = "integer", status: str = "normal"):
        self.value: int = value
        self.type: str = type
        self.status: str = status

NormalIntegerPropertyValue.get_classinfo().add_field('value', int)
NormalIntegerPropertyValue.get_classinfo().add_field('type', str, value_set(['integer']))
NormalIntegerPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalStringPropertyValue(_ACBaseType):
    """ A string property value containing a valid string.

    Attributes:
        value (:obj:`str`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: str, type: str = "string", status: str = "normal"):
        self.value: str = value
        self.type: str = type
        self.status: str = status

NormalStringPropertyValue.get_classinfo().add_field('value', str)
NormalStringPropertyValue.get_classinfo().add_field('type', str, value_set(['string']))
NormalStringPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalBooleanPropertyValue(_ACBaseType):
    """ A boolean property value containing a valid boolean value.

    Attributes:
        value (:obj:`bool`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: bool, type: str = "boolean", status: str = "normal"):
        self.value: bool = value
        self.type: str = type
        self.status: str = status

NormalBooleanPropertyValue.get_classinfo().add_field('value', bool)
NormalBooleanPropertyValue.get_classinfo().add_field('type', str, value_set(['boolean']))
NormalBooleanPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalLengthPropertyValue(_ACBaseType):
    """ A length property value containing a real length value. Value is measured in SI (meters).

    Attributes:
        value (:obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: float, type: str = "length", status: str = "normal"):
        self.value: float = value
        self.type: str = type
        self.status: str = status

NormalLengthPropertyValue.get_classinfo().add_field('value', float)
NormalLengthPropertyValue.get_classinfo().add_field('type', str, value_set(['length']))
NormalLengthPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalAreaPropertyValue(_ACBaseType):
    """ An area property value containing a real area. Value is measured in SI (square meters).

    Attributes:
        value (:obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: float, type: str = "area", status: str = "normal"):
        self.value: float = value
        self.type: str = type
        self.status: str = status

NormalAreaPropertyValue.get_classinfo().add_field('value', float)
NormalAreaPropertyValue.get_classinfo().add_field('type', str, value_set(['area']))
NormalAreaPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalVolumePropertyValue(_ACBaseType):
    """ A volume property value containing a real volume. Value is measured in SI (cubic meters).

    Attributes:
        value (:obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: float, type: str = "volume", status: str = "normal"):
        self.value: float = value
        self.type: str = type
        self.status: str = status

NormalVolumePropertyValue.get_classinfo().add_field('value', float)
NormalVolumePropertyValue.get_classinfo().add_field('type', str, value_set(['volume']))
NormalVolumePropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalAnglePropertyValue(_ACBaseType):
    """ An angle property value containing a real angle. Value is measured in SI (radians).

    Attributes:
        value (:obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: float, type: str = "angle", status: str = "normal"):
        self.value: float = value
        self.type: str = type
        self.status: str = status

NormalAnglePropertyValue.get_classinfo().add_field('value', float)
NormalAnglePropertyValue.get_classinfo().add_field('type', str, value_set(['angle']))
NormalAnglePropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalNumberListPropertyValue(_ACBaseType):
    """ A number list property value containing numbers in an array.

    Attributes:
        value (:obj:`list` of :obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[float], type: str = "numberList", status: str = "normal"):
        self.value: List[float] = value
        self.type: str = type
        self.status: str = status

NormalNumberListPropertyValue.get_classinfo().add_field('value', List[float])
NormalNumberListPropertyValue.get_classinfo().add_field('type', str, value_set(['numberList']))
NormalNumberListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalIntegerListPropertyValue(_ACBaseType):
    """ An integer list property value containing integers in an array.

    Attributes:
        value (:obj:`list` of :obj:`int`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[int], type: str = "integerList", status: str = "normal"):
        self.value: List[int] = value
        self.type: str = type
        self.status: str = status

NormalIntegerListPropertyValue.get_classinfo().add_field('value', List[int])
NormalIntegerListPropertyValue.get_classinfo().add_field('type', str, value_set(['integerList']))
NormalIntegerListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalStringListPropertyValue(_ACBaseType):
    """ A string list property value containing strings in an array.

    Attributes:
        value (:obj:`list` of :obj:`str`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[str], type: str = "stringList", status: str = "normal"):
        self.value: List[str] = value
        self.type: str = type
        self.status: str = status

NormalStringListPropertyValue.get_classinfo().add_field('value', List[str])
NormalStringListPropertyValue.get_classinfo().add_field('type', str, value_set(['stringList']))
NormalStringListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalBooleanListPropertyValue(_ACBaseType):
    """ A boolean list property value containing boolean values in an array.

    Attributes:
        value (:obj:`list` of :obj:`bool`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[bool], type: str = "booleanList", status: str = "normal"):
        self.value: List[bool] = value
        self.type: str = type
        self.status: str = status

NormalBooleanListPropertyValue.get_classinfo().add_field('value', List[bool])
NormalBooleanListPropertyValue.get_classinfo().add_field('type', str, value_set(['booleanList']))
NormalBooleanListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalLengthListPropertyValue(_ACBaseType):
    """ A length list property value containing length values in an array. Values are measured in SI (meters).

    Attributes:
        value (:obj:`list` of :obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[float], type: str = "lengthList", status: str = "normal"):
        self.value: List[float] = value
        self.type: str = type
        self.status: str = status

NormalLengthListPropertyValue.get_classinfo().add_field('value', List[float])
NormalLengthListPropertyValue.get_classinfo().add_field('type', str, value_set(['lengthList']))
NormalLengthListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalAreaListPropertyValue(_ACBaseType):
    """ A area list property value containing areas in an array. Values are measured in SI (square meters).

    Attributes:
        value (:obj:`list` of :obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[float], type: str = "areaList", status: str = "normal"):
        self.value: List[float] = value
        self.type: str = type
        self.status: str = status

NormalAreaListPropertyValue.get_classinfo().add_field('value', List[float])
NormalAreaListPropertyValue.get_classinfo().add_field('type', str, value_set(['areaList']))
NormalAreaListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalVolumeListPropertyValue(_ACBaseType):
    """ A volume list property value containing volumes in an array. Values are measured in SI (cubic meters).

    Attributes:
        value (:obj:`list` of :obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[float], type: str = "volumeList", status: str = "normal"):
        self.value: List[float] = value
        self.type: str = type
        self.status: str = status

NormalVolumeListPropertyValue.get_classinfo().add_field('value', List[float])
NormalVolumeListPropertyValue.get_classinfo().add_field('type', str, value_set(['volumeList']))
NormalVolumeListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalAngleListPropertyValue(_ACBaseType):
    """ A angle list property value containing angles in an array. Values are measured in SI (radians).

    Attributes:
        value (:obj:`list` of :obj:`float`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[float], type: str = "angleList", status: str = "normal"):
        self.value: List[float] = value
        self.type: str = type
        self.status: str = status

NormalAngleListPropertyValue.get_classinfo().add_field('value', List[float])
NormalAngleListPropertyValue.get_classinfo().add_field('type', str, value_set(['angleList']))
NormalAngleListPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class UserUndefinedPropertyValue(_ACBaseType):
    """ A userUndefined value means that there is no actual number/string/etc. value, but the user deliberately set an Undefined value: this is a valid value, too.

    Attributes:
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("type", "status", )

    def __init__(self, type: str, status: str = "userUndefined"):
        self.type: str = type
        self.status: str = status

UserUndefinedPropertyValue.get_classinfo().add_field('type', str, value_set(['number', 'integer', 'string', 'boolean', 'length', 'area', 'volume', 'angle', 'numberList', 'integerList', 'stringList', 'booleanList', 'lengthList', 'areaList', 'volumeList', 'angleList', 'singleEnum', 'multiEnum']))
UserUndefinedPropertyValue.get_classinfo().add_field('status', str, value_set(['userUndefined']))


class NotAvailablePropertyValue(_ACBaseType):
    """ A notAvailable value means that the property is not available for the property owner (and therefore it has no property value for it).

    Attributes:
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("type", "status", )

    def __init__(self, type: str, status: str = "notAvailable"):
        self.type: str = type
        self.status: str = status

NotAvailablePropertyValue.get_classinfo().add_field('type', str, value_set(['number', 'integer', 'string', 'boolean', 'length', 'area', 'volume', 'angle', 'numberList', 'integerList', 'stringList', 'booleanList', 'lengthList', 'areaList', 'volumeList', 'angleList', 'singleEnum', 'multiEnum']))
NotAvailablePropertyValue.get_classinfo().add_field('status', str, value_set(['notAvailable']))


class NotEvaluatedPropertyValue(_ACBaseType):
    """ A notEvaluated value means that the property could not be evaluated for the property owner for some reason.

    Attributes:
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("type", "status", )

    def __init__(self, type: str, status: str = "notEvaluated"):
        self.type: str = type
        self.status: str = status

NotEvaluatedPropertyValue.get_classinfo().add_field('type', str, value_set(['number', 'integer', 'string', 'boolean', 'length', 'area', 'volume', 'angle', 'numberList', 'integerList', 'stringList', 'booleanList', 'lengthList', 'areaList', 'volumeList', 'angleList', 'singleEnum', 'multiEnum']))
NotEvaluatedPropertyValue.get_classinfo().add_field('status', str, value_set(['notEvaluated']))


class DisplayValueEnumId(_ACBaseType):
    """ An enumeration value identifier using the displayed value.

    Attributes:
        displayValue (:obj:`str`): EMPTY STRING
        type (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("displayValue", "type", )

    def __init__(self, displayValue: str, type: str = "displayValue"):
        self.displayValue: str = displayValue
        self.type: str = type

DisplayValueEnumId.get_classinfo().add_field('displayValue', str)
DisplayValueEnumId.get_classinfo().add_field('type', str, value_set(['displayValue']))


EnumValueId = DisplayValueEnumId
""" The identifier of a property enumeration value.
"""


class PossibleEnumValue(_ACBaseType):
    """ Description of an enumeration value.

    Attributes:
        enumValueId (:obj:`EnumValueId`): The identifier of a property enumeration value.
        displayValue (:obj:`str`): Displayed value of the enumeration.

    """
    __slots__ = ("enumValueId", "displayValue", )

    def __init__(self, enumValueId: EnumValueId, displayValue: str):
        self.enumValueId: EnumValueId = enumValueId
        self.displayValue: str = displayValue

PossibleEnumValue.get_classinfo().add_field('enumValueId', EnumValueId)
PossibleEnumValue.get_classinfo().add_field('displayValue', str)


class PossibleEnumValuesArrayItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        enumValue (:obj:`PossibleEnumValue`): Description of an enumeration value.

    """
    __slots__ = ("enumValue", )

    def __init__(self, enumValue: PossibleEnumValue):
        self.enumValue: PossibleEnumValue = enumValue

PossibleEnumValuesArrayItem.get_classinfo().add_field('enumValue', PossibleEnumValue)


class Error(_ACBaseType):
    """ Error details.

    Attributes:
        code (:obj:`int`): The error code.
        message (:obj:`str`): The error message.

    """
    __slots__ = ("code", "message", )

    def __init__(self, code: int, message: str):
        self.code: int = code
        self.message: str = message

Error.get_classinfo().add_field('code', int)
Error.get_classinfo().add_field('message', str)


class ErrorItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        error (:obj:`Error`): Error details.

    """
    __slots__ = ("error", )

    def __init__(self, error: Error):
        self.error: Error = error

ErrorItem.get_classinfo().add_field('error', Error)


class SuccessfulExecutionResult(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        success (:obj:`bool`): EMPTY STRING

    """
    __slots__ = ("success", )

    def __init__(self, success: bool):
        self.success: bool = success

SuccessfulExecutionResult.get_classinfo().add_field('success', bool)


class FailedExecutionResult(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        success (:obj:`bool`): EMPTY STRING
        error (:obj:`Error`): Error details.

    """
    __slots__ = ("success", "error", )

    def __init__(self, success: bool, error: Error):
        self.success: bool = success
        self.error: Error = error

FailedExecutionResult.get_classinfo().add_field('success', bool)
FailedExecutionResult.get_classinfo().add_field('error', Error)


class ExecutionResult(_ACUnionType):
    """ The result of the execution for one case.

    Attributes:
        success (:obj:`bool`): None
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("success", "error", )

    constructor  = _ConstructUnion(Union[SuccessfulExecutionResult, FailedExecutionResult])

    def __new__(cls, success: bool, error: Optional[Error] = None):
        return cls.constructor(success=success, error=error)

    def __init__(self, success: bool, error: Optional[Error] = None):
        self.success: bool = success
        self.error: Optional[Error] = error


class ElementId(_ACBaseType):
    """ The identifier of an element.

    Attributes:
        guid (:obj:`UUID`): A Globally Unique Identifier (or Universally Unique Identifier) in its string representation as defined in RFC 4122.

    """
    __slots__ = ("guid", )

    def __init__(self, guid: UUID):
        self.guid: UUID = guid

ElementId.get_classinfo().add_field('guid', UUID)


class ElementIdArrayItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        elementId (:obj:`ElementId`): The identifier of an element.

    """
    __slots__ = ("elementId", )

    def __init__(self, elementId: ElementId):
        self.elementId: ElementId = elementId

ElementIdArrayItem.get_classinfo().add_field('elementId', ElementId)


class ElementsWrapper(_ACBaseType):
    """ List of elements.

    Attributes:
        elements (:obj:`list` of :obj:`ElementIdArrayItem`): List of the elements.

    """
    __slots__ = ("elements", )

    def __init__(self, elements: List[ElementIdArrayItem]):
        self.elements: List[ElementIdArrayItem] = elements

ElementsWrapper.get_classinfo().add_field('elements', List[ElementIdArrayItem])


class ElementsOrError(_ACUnionType):
    """ List of elements or error.

    Attributes:
        elements (:obj:`list` of :obj:`ElementIdArrayItem`, optional): List of the elements.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("elements", "error", )

    constructor  = _ConstructUnion(Union[ElementsWrapper, ErrorItem])

    def __new__(cls, elements: Optional[List[ElementIdArrayItem]] = None, error: Optional[Error] = None):
        return cls.constructor(elements=elements, error=error)

    def __init__(self, elements: Optional[List[ElementIdArrayItem]] = None, error: Optional[Error] = None):
        self.elements: Optional[List[ElementIdArrayItem]] = elements
        self.error: Optional[Error] = error


class NavigatorItemId(_ACBaseType):
    """ The identifier of a navigator item.

    Attributes:
        guid (:obj:`UUID`): A Globally Unique Identifier (or Universally Unique Identifier) in its string representation as defined in RFC 4122.

    """
    __slots__ = ("guid", )

    def __init__(self, guid: UUID):
        self.guid: UUID = guid

NavigatorItemId.get_classinfo().add_field('guid', UUID)


class PublisherSetId(_ACBaseType):
    """ The identifier of a publisher set.

    Attributes:
        name (:obj:`str`): Name of the publisher set.
        type (:obj:`str`): Type of the navigator item tree.

    """
    __slots__ = ("name", "type", )

    def __init__(self, name: str, type: str = "PublisherSets"):
        self.name: str = name
        self.type: str = type

PublisherSetId.get_classinfo().add_field('name', str)
PublisherSetId.get_classinfo().add_field('type', str, value_set(['PublisherSets']))


class OtherNavigatorTreeId(_ACBaseType):
    """ The identifier of a navigator item tree.

    Attributes:
        type (:obj:`str`): Type of the navigator item tree.

    """
    __slots__ = ("type", )

    def __init__(self, type: str):
        self.type: str = type

OtherNavigatorTreeId.get_classinfo().add_field('type', str, value_set(['ProjectMap', 'ViewMap', 'MyViewMap', 'LayoutBook']))


class FolderParameters(_ACBaseType):
    """ The parameters of a folder.

    Attributes:
        name (:obj:`str`): The name of the folder to create.

    """
    __slots__ = ("name", )

    def __init__(self, name: str):
        self.name: str = name

FolderParameters.get_classinfo().add_field('name', str)


class NavigatorTreeId(_ACUnionType):
    """ The identifier of a navigator item tree.

    Attributes:
        type (:obj:`str`): Type of the navigator item tree.
        name (:obj:`str`, optional): Name of the publisher set.

    """
    __slots__ = ("type", "name", )

    constructor  = _ConstructUnion(Union[PublisherSetId, OtherNavigatorTreeId])

    def __new__(cls, type: str, name: Optional[str] = None):
        return cls.constructor(type=type, name=name)

    def __init__(self, type: str, name: Optional[str] = None):
        self.type: str = type
        self.name: Optional[str] = name


class BoundingBox2D(_ACBaseType):
    """ 2D bounding box of an element.

    Attributes:
        xMin (:obj:`float`): Minimum X value of bounding box.
        yMin (:obj:`float`): Minimum Y value of bounding box.
        xMax (:obj:`float`): Maximum X value of bounding box.
        yMax (:obj:`float`): Maximum Y value of bounding box.

    """
    __slots__ = ("xMin", "yMin", "xMax", "yMax", )

    def __init__(self, xMin: float, yMin: float, xMax: float, yMax: float):
        self.xMin: float = xMin
        self.yMin: float = yMin
        self.xMax: float = xMax
        self.yMax: float = yMax

BoundingBox2D.get_classinfo().add_field('xMin', float)
BoundingBox2D.get_classinfo().add_field('yMin', float)
BoundingBox2D.get_classinfo().add_field('xMax', float)
BoundingBox2D.get_classinfo().add_field('yMax', float)


class BoundingBox3D(_ACBaseType):
    """ 3D bounding box of an element.

    Attributes:
        xMin (:obj:`float`): Minimum X value of bounding box.
        yMin (:obj:`float`): Minimum Y value of bounding box.
        zMin (:obj:`float`): Minimum Z value of bounding box.
        xMax (:obj:`float`): Maximum X value of bounding box.
        yMax (:obj:`float`): Maximum Y value of bounding box.
        zMax (:obj:`float`): Maximum Z value of bounding box.

    """
    __slots__ = ("xMin", "yMin", "zMin", "xMax", "yMax", "zMax", )

    def __init__(self, xMin: float, yMin: float, zMin: float, xMax: float, yMax: float, zMax: float):
        self.xMin: float = xMin
        self.yMin: float = yMin
        self.zMin: float = zMin
        self.xMax: float = xMax
        self.yMax: float = yMax
        self.zMax: float = zMax

BoundingBox3D.get_classinfo().add_field('xMin', float)
BoundingBox3D.get_classinfo().add_field('yMin', float)
BoundingBox3D.get_classinfo().add_field('zMin', float)
BoundingBox3D.get_classinfo().add_field('xMax', float)
BoundingBox3D.get_classinfo().add_field('yMax', float)
BoundingBox3D.get_classinfo().add_field('zMax', float)


class Subset(_ACBaseType):
    """ Provides a way to assign IDs to the layouts contained in the subset.

    Attributes:
        name (:obj:`str`): Defines the name for the layout subset.
        includeToIDSequence (:obj:`bool`): Indicates whether this subset is included in or excluded from automatic ID assignment.
        customNumbering (:obj:`bool`): Defines whether the IDs are generated automatically or use a custom numbering.
        continueNumbering (:obj:`bool`): Continue using the ID assignment of upper levels. Layouts within this subset are going to be assigned IDs as if they were not within this subset, but part of the level above. In this case you are only using the Subset as a logical grouping which has no effect on IDs.
        useUpperPrefix (:obj:`bool`): Use the prefix and ID of upper levels. Layouts in this subset will be assigned IDs based on the previous layout in the layout book structure.
        addOwnPrefix (:obj:`bool`): Adds own prefix to the subset.
        customNumber (:obj:`str`): Specifies the custom subset ID.
        autoNumber (:obj:`str`): Specifies the automatic subset ID.
        numberingStyle (:obj:`str`): List of the supported numbering styles.
        startAt (:obj:`int`): Specifies a number from which the numbering starts.
        ownPrefix (:obj:`str`): Defines a custom prefix for the subset.

    """
    __slots__ = ("name", "includeToIDSequence", "customNumbering", "continueNumbering", "useUpperPrefix", "addOwnPrefix", "customNumber", "autoNumber", "numberingStyle", "startAt", "ownPrefix", )

    def __init__(self, name: str, includeToIDSequence: bool, customNumbering: bool, continueNumbering: bool, useUpperPrefix: bool, addOwnPrefix: bool, customNumber: str, autoNumber: str, numberingStyle: str, startAt: int, ownPrefix: str):
        self.name: str = name
        self.includeToIDSequence: bool = includeToIDSequence
        self.customNumbering: bool = customNumbering
        self.continueNumbering: bool = continueNumbering
        self.useUpperPrefix: bool = useUpperPrefix
        self.addOwnPrefix: bool = addOwnPrefix
        self.customNumber: str = customNumber
        self.autoNumber: str = autoNumber
        self.numberingStyle: str = numberingStyle
        self.startAt: int = startAt
        self.ownPrefix: str = ownPrefix

Subset.get_classinfo().add_field('name', str, min_length(1))
Subset.get_classinfo().add_field('includeToIDSequence', bool)
Subset.get_classinfo().add_field('customNumbering', bool)
Subset.get_classinfo().add_field('continueNumbering', bool)
Subset.get_classinfo().add_field('useUpperPrefix', bool)
Subset.get_classinfo().add_field('addOwnPrefix', bool)
Subset.get_classinfo().add_field('customNumber', str)
Subset.get_classinfo().add_field('autoNumber', str)
Subset.get_classinfo().add_field('numberingStyle', str, value_set(['Undefined', 'abc', 'ABC', '1', '01', '001', '0001', 'noID']))
Subset.get_classinfo().add_field('startAt', int)
Subset.get_classinfo().add_field('ownPrefix', str)


class LayoutParameters(_ACBaseType):
    """ The parameters of the layout.

    Attributes:
        horizontalSize (:obj:`float`): Horizontal size of the layout in millimeters.
        verticalSize (:obj:`float`): Vertical size of the layout in millimeters.
        leftMargin (:obj:`float`): Layout margin from the left side of the paper.
        topMargin (:obj:`float`): Layout margin from the top side of the paper.
        rightMargin (:obj:`float`): Layout margin from the right side of the paper.
        bottomMargin (:obj:`float`): Layout margin from the bottom side of the paper.
        customLayoutNumber (:obj:`str`): Specifies the custom ID.
        customLayoutNumbering (:obj:`bool`): Defines whether a unique ID is used for the current layout or not.
        doNotIncludeInNumbering (:obj:`bool`): Indicates whether this layout is included in or excluded from automatic ID assignment.
        displayMasterLayoutBelow (:obj:`bool`): Display the master layout above or below the layout.
        layoutPageNumber (:obj:`int`): Page number of layout when this layout contains multi-page drawings.
        actPageIndex (:obj:`int`): The actual index of layout inside the multi-page layout.
        currentRevisionId (:obj:`str`): ID of the current document revision of the layout.
        currentFinalRevisionId (:obj:`str`): ID with optional suffix of the current document revision of the layout.
        hasIssuedRevision (:obj:`bool`): One or more issued document revisions have already been created for the layout.
        hasActualRevision (:obj:`bool`): An open document revision exists for the layout.

    """
    __slots__ = ("horizontalSize", "verticalSize", "leftMargin", "topMargin", "rightMargin", "bottomMargin", "customLayoutNumber", "customLayoutNumbering", "doNotIncludeInNumbering", "displayMasterLayoutBelow", "layoutPageNumber", "actPageIndex", "currentRevisionId", "currentFinalRevisionId", "hasIssuedRevision", "hasActualRevision", )

    def __init__(self, horizontalSize: float, verticalSize: float, leftMargin: float, topMargin: float, rightMargin: float, bottomMargin: float, customLayoutNumber: str, customLayoutNumbering: bool, doNotIncludeInNumbering: bool, displayMasterLayoutBelow: bool, layoutPageNumber: int, actPageIndex: int, currentRevisionId: str, currentFinalRevisionId: str, hasIssuedRevision: bool, hasActualRevision: bool):
        self.horizontalSize: float = horizontalSize
        self.verticalSize: float = verticalSize
        self.leftMargin: float = leftMargin
        self.topMargin: float = topMargin
        self.rightMargin: float = rightMargin
        self.bottomMargin: float = bottomMargin
        self.customLayoutNumber: str = customLayoutNumber
        self.customLayoutNumbering: bool = customLayoutNumbering
        self.doNotIncludeInNumbering: bool = doNotIncludeInNumbering
        self.displayMasterLayoutBelow: bool = displayMasterLayoutBelow
        self.layoutPageNumber: int = layoutPageNumber
        self.actPageIndex: int = actPageIndex
        self.currentRevisionId: str = currentRevisionId
        self.currentFinalRevisionId: str = currentFinalRevisionId
        self.hasIssuedRevision: bool = hasIssuedRevision
        self.hasActualRevision: bool = hasActualRevision

LayoutParameters.get_classinfo().add_field('horizontalSize', float)
LayoutParameters.get_classinfo().add_field('verticalSize', float)
LayoutParameters.get_classinfo().add_field('leftMargin', float)
LayoutParameters.get_classinfo().add_field('topMargin', float)
LayoutParameters.get_classinfo().add_field('rightMargin', float)
LayoutParameters.get_classinfo().add_field('bottomMargin', float)
LayoutParameters.get_classinfo().add_field('customLayoutNumber', str)
LayoutParameters.get_classinfo().add_field('customLayoutNumbering', bool)
LayoutParameters.get_classinfo().add_field('doNotIncludeInNumbering', bool)
LayoutParameters.get_classinfo().add_field('displayMasterLayoutBelow', bool)
LayoutParameters.get_classinfo().add_field('layoutPageNumber', int)
LayoutParameters.get_classinfo().add_field('actPageIndex', int)
LayoutParameters.get_classinfo().add_field('currentRevisionId', str)
LayoutParameters.get_classinfo().add_field('currentFinalRevisionId', str)
LayoutParameters.get_classinfo().add_field('hasIssuedRevision', bool)
LayoutParameters.get_classinfo().add_field('hasActualRevision', bool)


class ClassificationIdWrapper(_ACBaseType):
    """ 

    Attributes:
        classificationId (:obj:`ClassificationId`): The element classification identifier.

    """
    __slots__ = ("classificationId", )

    def __init__(self, classificationId: ClassificationId):
        self.classificationId: ClassificationId = classificationId

ClassificationIdWrapper.get_classinfo().add_field('classificationId', ClassificationId)


class ClassificationItemDetailsWrapper(_ACBaseType):
    """ 

    Attributes:
        classificationItem (:obj:`ClassificationItemDetails`): Details of a classification item.

    """
    __slots__ = ("classificationItem", )

    def __init__(self, classificationItem: ClassificationItemDetails):
        self.classificationItem: ClassificationItemDetails = classificationItem

ClassificationItemDetailsWrapper.get_classinfo().add_field('classificationItem', ClassificationItemDetails)


class EnumValueIdWrapper(_ACBaseType):
    """ 

    Attributes:
        enumValueId (:obj:`EnumValueId`): The identifier of a property enumeration value.

    """
    __slots__ = ("enumValueId", )

    def __init__(self, enumValueId: EnumValueId):
        self.enumValueId: EnumValueId = enumValueId

EnumValueIdWrapper.get_classinfo().add_field('enumValueId', EnumValueId)


class NavigatorItemIdWrapper(_ACBaseType):
    """ 

    Attributes:
        navigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.

    """
    __slots__ = ("navigatorItemId", )

    def __init__(self, navigatorItemId: NavigatorItemId):
        self.navigatorItemId: NavigatorItemId = navigatorItemId

NavigatorItemIdWrapper.get_classinfo().add_field('navigatorItemId', NavigatorItemId)


class BoundingBox2DWrapper(_ACBaseType):
    """ 

    Attributes:
        boundingBox2D (:obj:`BoundingBox2D`): 2D bounding box of an element.

    """
    __slots__ = ("boundingBox2D", )

    def __init__(self, boundingBox2D: BoundingBox2D):
        self.boundingBox2D: BoundingBox2D = boundingBox2D

BoundingBox2DWrapper.get_classinfo().add_field('boundingBox2D', BoundingBox2D)


class BoundingBox3DWrapper(_ACBaseType):
    """ 

    Attributes:
        boundingBox3D (:obj:`BoundingBox3D`): 3D bounding box of an element.

    """
    __slots__ = ("boundingBox3D", )

    def __init__(self, boundingBox3D: BoundingBox3D):
        self.boundingBox3D: BoundingBox3D = boundingBox3D

BoundingBox3DWrapper.get_classinfo().add_field('boundingBox3D', BoundingBox3D)


class ClassificationIdOrError(_ACUnionType):
    """ EMPTY STRING

    Attributes:
        classificationId (:obj:`ClassificationId`, optional): The element classification identifier.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("classificationId", "error", )

    constructor  = _ConstructUnion(Union[ClassificationIdWrapper, ErrorItem])

    def __new__(cls, classificationId: Optional[ClassificationId] = None, error: Optional[Error] = None):
        return cls.constructor(classificationId=classificationId, error=error)

    def __init__(self, classificationId: Optional[ClassificationId] = None, error: Optional[Error] = None):
        self.classificationId: Optional[ClassificationId] = classificationId
        self.error: Optional[Error] = error


class ElementClassification(_ACBaseType):
    """ Element classification.

    Attributes:
        elementId (:obj:`ElementId`): The identifier of an element.
        classificationId (:obj:`ClassificationId`): The element classification identifier.

    """
    __slots__ = ("elementId", "classificationId", )

    def __init__(self, elementId: ElementId, classificationId: ClassificationId):
        self.elementId: ElementId = elementId
        self.classificationId: ClassificationId = classificationId

ElementClassification.get_classinfo().add_field('elementId', ElementId)
ElementClassification.get_classinfo().add_field('classificationId', ClassificationId)


class ClassificationItemOrError(_ACUnionType):
    """ EMPTY STRING

    Attributes:
        classificationItem (:obj:`ClassificationItemDetails`, optional): Details of a classification item.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("classificationItem", "error", )

    constructor  = _ConstructUnion(Union[ClassificationItemDetailsWrapper, ErrorItem])

    def __new__(cls, classificationItem: Optional[ClassificationItemDetails] = None, error: Optional[Error] = None):
        return cls.constructor(classificationItem=classificationItem, error=error)

    def __init__(self, classificationItem: Optional[ClassificationItemDetails] = None, error: Optional[Error] = None):
        self.classificationItem: Optional[ClassificationItemDetails] = classificationItem
        self.error: Optional[Error] = error


class PropertyIdOrError(_ACUnionType):
    """ A property identifier or error.

    Attributes:
        propertyId (:obj:`PropertyId`, optional): The identifier of a property.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("propertyId", "error", )

    constructor  = _ConstructUnion(Union[PropertyIdArrayItem, ErrorItem])

    def __new__(cls, propertyId: Optional[PropertyId] = None, error: Optional[Error] = None):
        return cls.constructor(propertyId=propertyId, error=error)

    def __init__(self, propertyId: Optional[PropertyId] = None, error: Optional[Error] = None):
        self.propertyId: Optional[PropertyId] = propertyId
        self.error: Optional[Error] = error


class PropertyDefinition(_ACBaseType):
    """ A property definition.

    Attributes:
        group (:obj:`PropertyGroup`): A property group.
        name (:obj:`str`): The localized name of the property.
        description (:obj:`str`): The description of the property.
        possibleEnumValues (:obj:`list` of :obj:`PossibleEnumValuesArrayItem`, optional): List of enumeration values.

    """
    __slots__ = ("group", "name", "description", "possibleEnumValues", )

    def __init__(self, group: PropertyGroup, name: str, description: str, possibleEnumValues: Optional[List[PossibleEnumValuesArrayItem]] = None):
        self.group: PropertyGroup = group
        self.name: str = name
        self.description: str = description
        self.possibleEnumValues: Optional[List[PossibleEnumValuesArrayItem]] = possibleEnumValues

PropertyDefinition.get_classinfo().add_field('group', PropertyGroup)
PropertyDefinition.get_classinfo().add_field('name', str)
PropertyDefinition.get_classinfo().add_field('description', str)
PropertyDefinition.get_classinfo().add_field('possibleEnumValues', Optional[List[PossibleEnumValuesArrayItem]])


class NormalSingleEnumPropertyValue(_ACBaseType):
    """ A single enumeration property value containing the ID of the selected enum value.

    Attributes:
        value (:obj:`EnumValueId`): The identifier of a property enumeration value.
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: EnumValueId, type: str = "singleEnum", status: str = "normal"):
        self.value: EnumValueId = value
        self.type: str = type
        self.status: str = status

NormalSingleEnumPropertyValue.get_classinfo().add_field('value', EnumValueId)
NormalSingleEnumPropertyValue.get_classinfo().add_field('type', str, value_set(['singleEnum']))
NormalSingleEnumPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class NormalMultiEnumPropertyValue(_ACBaseType):
    """ A multiple choice enumeration property value containing the IDs of the selected enum values in an array.

    Attributes:
        value (:obj:`list` of :obj:`EnumValueIdWrapper`): List of enumeration identifiers.
        type (:obj:`str`): EMPTY STRING
        status (:obj:`str`): EMPTY STRING

    """
    __slots__ = ("value", "type", "status", )

    def __init__(self, value: List[EnumValueIdWrapper], type: str = "multiEnum", status: str = "normal"):
        self.value: List[EnumValueIdWrapper] = value
        self.type: str = type
        self.status: str = status

NormalMultiEnumPropertyValue.get_classinfo().add_field('value', List[EnumValueIdWrapper])
NormalMultiEnumPropertyValue.get_classinfo().add_field('type', str, value_set(['multiEnum']))
NormalMultiEnumPropertyValue.get_classinfo().add_field('status', str, value_set(['normal']))


class BoundingBox2DOrError(_ACUnionType):
    """ A 2D bounding box or error.

    Attributes:
        boundingBox2D (:obj:`BoundingBox2D`, optional): 2D bounding box of an element.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("boundingBox2D", "error", )

    constructor  = _ConstructUnion(Union[BoundingBox2DWrapper, ErrorItem])

    def __new__(cls, boundingBox2D: Optional[BoundingBox2D] = None, error: Optional[Error] = None):
        return cls.constructor(boundingBox2D=boundingBox2D, error=error)

    def __init__(self, boundingBox2D: Optional[BoundingBox2D] = None, error: Optional[Error] = None):
        self.boundingBox2D: Optional[BoundingBox2D] = boundingBox2D
        self.error: Optional[Error] = error


class BoundingBox3DOrError(_ACUnionType):
    """ A 3D bounding box or error.

    Attributes:
        boundingBox3D (:obj:`BoundingBox3D`, optional): 3D bounding box of an element.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("boundingBox3D", "error", )

    constructor  = _ConstructUnion(Union[BoundingBox3DWrapper, ErrorItem])

    def __new__(cls, boundingBox3D: Optional[BoundingBox3D] = None, error: Optional[Error] = None):
        return cls.constructor(boundingBox3D=boundingBox3D, error=error)

    def __init__(self, boundingBox3D: Optional[BoundingBox3D] = None, error: Optional[Error] = None):
        self.boundingBox3D: Optional[BoundingBox3D] = boundingBox3D
        self.error: Optional[Error] = error


class ClassificationIdsOrErrorsWrapper(_ACBaseType):
    """ 

    Attributes:
        classificationIds (:obj:`list` of :obj:`ClassificationIdOrError`): The list of element classification identifiers or errors.

    """
    __slots__ = ("classificationIds", )

    def __init__(self, classificationIds: List[ClassificationIdOrError]):
        self.classificationIds: List[ClassificationIdOrError] = classificationIds

ClassificationIdsOrErrorsWrapper.get_classinfo().add_field('classificationIds', List[ClassificationIdOrError])


class PropertyDefinitionWrapper(_ACBaseType):
    """ 

    Attributes:
        propertyDefinition (:obj:`PropertyDefinition`): A property definition.

    """
    __slots__ = ("propertyDefinition", )

    def __init__(self, propertyDefinition: PropertyDefinition):
        self.propertyDefinition: PropertyDefinition = propertyDefinition

PropertyDefinitionWrapper.get_classinfo().add_field('propertyDefinition', PropertyDefinition)


class ElementClassificationOrError(_ACUnionType):
    """ Element classification identifiers or error.

    Attributes:
        classificationIds (:obj:`list` of :obj:`ClassificationIdOrError`, optional): The list of element classification identifiers or errors.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("classificationIds", "error", )

    constructor  = _ConstructUnion(Union[ClassificationIdsOrErrorsWrapper, ErrorItem])

    def __new__(cls, classificationIds: Optional[List[ClassificationIdOrError]] = None, error: Optional[Error] = None):
        return cls.constructor(classificationIds=classificationIds, error=error)

    def __init__(self, classificationIds: Optional[List[ClassificationIdOrError]] = None, error: Optional[Error] = None):
        self.classificationIds: Optional[List[ClassificationIdOrError]] = classificationIds
        self.error: Optional[Error] = error


class PropertyDefinitionOrError(_ACUnionType):
    """ A property definition or error.

    Attributes:
        propertyDefinition (:obj:`PropertyDefinition`, optional): A property definition.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("propertyDefinition", "error", )

    constructor  = _ConstructUnion(Union[PropertyDefinitionWrapper, ErrorItem])

    def __new__(cls, propertyDefinition: Optional[PropertyDefinition] = None, error: Optional[Error] = None):
        return cls.constructor(propertyDefinition=propertyDefinition, error=error)

    def __init__(self, propertyDefinition: Optional[PropertyDefinition] = None, error: Optional[Error] = None):
        self.propertyDefinition: Optional[PropertyDefinition] = propertyDefinition
        self.error: Optional[Error] = error


class ClassificationItemInTree_: pass
class ClassificationItemArrayItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        classificationItem (:obj:`ClassificationItemInTree_`): Details of a classification item.

    """
    __slots__ = ("classificationItem", )

    def __init__(self, classificationItem: ClassificationItemInTree_):
        self.classificationItem: ClassificationItemInTree_ = classificationItem


class ClassificationItemInTree(_ACBaseType):
    """ Details of a classification item.

    Attributes:
        classificationItemId (:obj:`ClassificationItemId`): The identifier of a classification item.
        id (:obj:`str`): The user specified unique identifier of the classification item.
        name (:obj:`str`): The display name of the classification item.
        description (:obj:`str`): The description of the classification item.
        children (:obj:`list` of :obj:`ClassificationItemArrayItem`, optional): The list of classification items.

    """
    __slots__ = ("classificationItemId", "id", "name", "description", "children", )

    def __init__(self, classificationItemId: ClassificationItemId, id: str, name: str, description: str, children: Optional[List[ClassificationItemArrayItem]] = None):
        self.classificationItemId: ClassificationItemId = classificationItemId
        self.id: str = id
        self.name: str = name
        self.description: str = description
        self.children: Optional[List[ClassificationItemArrayItem]] = children

ClassificationItemInTree.get_classinfo().add_field('classificationItemId', ClassificationItemId)
ClassificationItemInTree.get_classinfo().add_field('id', str)
ClassificationItemInTree.get_classinfo().add_field('name', str)
ClassificationItemInTree.get_classinfo().add_field('description', str)
ClassificationItemInTree.get_classinfo().add_field('children', Optional[List[ClassificationItemArrayItem]])

ClassificationItemInTree_ = ClassificationItemInTree
ClassificationItemArrayItem.get_classinfo().add_field('classificationItem', ClassificationItemInTree)


class NormalOrUserUndefinedPropertyValue(_ACUnionType):
    """ A normal or a userUndefined property value.

    Attributes:
        type (:obj:`str`): None
        status (:obj:`str`): None
        value (:obj:`float`, :obj:`int`, :obj:`str`, :obj:`bool`, :obj:`list` of :obj:`float`, :obj:`list` of :obj:`int`, :obj:`list` of :obj:`str`, :obj:`list` of :obj:`bool`, :obj:`EnumValueId`, :obj:`list` of :obj:`EnumValueIdWrapper`): None; The identifier of a property enumeration value.; List of enumeration identifiers.

    """
    __slots__ = ("type", "status", "value", )

    constructor  = _ConstructUnion(Union[NormalNumberPropertyValue, NormalIntegerPropertyValue, NormalStringPropertyValue, NormalBooleanPropertyValue, NormalLengthPropertyValue, NormalAreaPropertyValue, NormalVolumePropertyValue, NormalAnglePropertyValue, NormalNumberListPropertyValue, NormalIntegerListPropertyValue, NormalStringListPropertyValue, NormalBooleanListPropertyValue, NormalLengthListPropertyValue, NormalAreaListPropertyValue, NormalVolumeListPropertyValue, NormalAngleListPropertyValue, NormalSingleEnumPropertyValue, NormalMultiEnumPropertyValue, UserUndefinedPropertyValue])

    def __new__(cls, type: str, status: str, value: Union[float, int, str, bool, List[float], List[int], List[str], List[bool], EnumValueId, List[EnumValueIdWrapper], None] = None):
        return cls.constructor(type=type, status=status, value=value)

    def __init__(self, type: str, status: str, value: Union[float, int, str, bool, List[float], List[int], List[str], List[bool], EnumValueId, List[EnumValueIdWrapper], None] = None):
        self.type: str = type
        self.status: str = status
        self.value: Union[float, int, str, bool, List[float], List[int], List[str], List[bool], EnumValueId, List[EnumValueIdWrapper], None] = value


class ElementPropertyValue(_ACBaseType):
    """ A property value with the identifier of the property and the owner element.

    Attributes:
        elementId (:obj:`ElementId`): The identifier of an element.
        propertyId (:obj:`PropertyId`): The identifier of a property.
        propertyValue (:obj:`NormalOrUserUndefinedPropertyValue`): A normal or a userUndefined property value.

    """
    __slots__ = ("elementId", "propertyId", "propertyValue", )

    def __init__(self, elementId: ElementId, propertyId: PropertyId, propertyValue: NormalOrUserUndefinedPropertyValue):
        self.elementId: ElementId = elementId
        self.propertyId: PropertyId = propertyId
        self.propertyValue: NormalOrUserUndefinedPropertyValue = propertyValue

ElementPropertyValue.get_classinfo().add_field('elementId', ElementId)
ElementPropertyValue.get_classinfo().add_field('propertyId', PropertyId)
ElementPropertyValue.get_classinfo().add_field('propertyValue', NormalOrUserUndefinedPropertyValue)


class PropertyValue(_ACUnionType):
    """ A normal, userUndefined, notAvailable or notEvaluated property value.

    Attributes:
        type (:obj:`str`): None
        status (:obj:`str`): None
        value (:obj:`float`, :obj:`int`, :obj:`str`, :obj:`bool`, :obj:`list` of :obj:`float`, :obj:`list` of :obj:`int`, :obj:`list` of :obj:`str`, :obj:`list` of :obj:`bool`, :obj:`EnumValueId`, :obj:`list` of :obj:`EnumValueIdWrapper`): None; The identifier of a property enumeration value.; List of enumeration identifiers.

    """
    __slots__ = ("type", "status", "value", )

    constructor  = _ConstructUnion(Union[NormalOrUserUndefinedPropertyValue, NotAvailablePropertyValue, NotEvaluatedPropertyValue])

    def __new__(cls, type: str, status: str, value: Union[float, int, str, bool, List[float], List[int], List[str], List[bool], EnumValueId, List[EnumValueIdWrapper], None] = None):
        return cls.constructor(type=type, status=status, value=value)

    def __init__(self, type: str, status: str, value: Union[float, int, str, bool, List[float], List[int], List[str], List[bool], EnumValueId, List[EnumValueIdWrapper], None] = None):
        self.type: str = type
        self.status: str = status
        self.value: Union[float, int, str, bool, List[float], List[int], List[str], List[bool], EnumValueId, List[EnumValueIdWrapper], None] = value


class PropertyValueWrapper(_ACBaseType):
    """ 

    Attributes:
        propertyValue (:obj:`PropertyValue`): A normal, userUndefined, notAvailable or notEvaluated property value.

    """
    __slots__ = ("propertyValue", )

    def __init__(self, propertyValue: PropertyValue):
        self.propertyValue: PropertyValue = propertyValue

PropertyValueWrapper.get_classinfo().add_field('propertyValue', PropertyValue)


class PropertyValueOrErrorItem(_ACUnionType):
    """ Contains a property value or an error for the property.

    Attributes:
        propertyValue (:obj:`PropertyValue`, optional): A normal, userUndefined, notAvailable or notEvaluated property value.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("propertyValue", "error", )

    constructor  = _ConstructUnion(Union[PropertyValueWrapper, ErrorItem])

    def __new__(cls, propertyValue: Optional[PropertyValue] = None, error: Optional[Error] = None):
        return cls.constructor(propertyValue=propertyValue, error=error)

    def __init__(self, propertyValue: Optional[PropertyValue] = None, error: Optional[Error] = None):
        self.propertyValue: Optional[PropertyValue] = propertyValue
        self.error: Optional[Error] = error


class PropertyValuesWrapper(_ACBaseType):
    """ 

    Attributes:
        propertyValues (:obj:`list` of :obj:`PropertyValueOrErrorItem`): List of property values.

    """
    __slots__ = ("propertyValues", )

    def __init__(self, propertyValues: List[PropertyValueOrErrorItem]):
        self.propertyValues: List[PropertyValueOrErrorItem] = propertyValues

PropertyValuesWrapper.get_classinfo().add_field('propertyValues', List[PropertyValueOrErrorItem])


class PropertyValuesOrError(_ACUnionType):
    """ Contains either a list of property values or an error.

    Attributes:
        propertyValues (:obj:`list` of :obj:`PropertyValueOrErrorItem`, optional): List of property values.
        error (:obj:`Error`, optional): Error details.

    """
    __slots__ = ("propertyValues", "error", )

    constructor  = _ConstructUnion(Union[PropertyValuesWrapper, ErrorItem])

    def __new__(cls, propertyValues: Optional[List[PropertyValueOrErrorItem]] = None, error: Optional[Error] = None):
        return cls.constructor(propertyValues=propertyValues, error=error)

    def __init__(self, propertyValues: Optional[List[PropertyValueOrErrorItem]] = None, error: Optional[Error] = None):
        self.propertyValues: Optional[List[PropertyValueOrErrorItem]] = propertyValues
        self.error: Optional[Error] = error


class NavigatorItem_: pass
class NavigatorItemArrayItem(_ACBaseType):
    """ EMPTY STRING

    Attributes:
        navigatorItem (:obj:`NavigatorItem_`): Details of a navigator item.

    """
    __slots__ = ("navigatorItem", )

    def __init__(self, navigatorItem: NavigatorItem_):
        self.navigatorItem: NavigatorItem_ = navigatorItem


class NavigatorItem(_ACBaseType):
    """ Details of a navigator item.

    Attributes:
        navigatorItemId (:obj:`NavigatorItemId`): The identifier of a navigator item.
        prefix (:obj:`str`): Prefix of the name of the navigator item.
        name (:obj:`str`): Name of the navigator item.
        type (:obj:`str`): Possible types of a navigator item. Most of the names are self-explanatory, the only exception is 'UndefinedItem', which means that the type of this navigator item cannot be retrieved from ARCHICAD yet.
        sourceNavigatorItemId (:obj:`NavigatorItemId`, optional): The identifier of a navigator item.
        children (:obj:`list` of :obj:`NavigatorItemArrayItem`, optional): List of navigator items.

    """
    __slots__ = ("navigatorItemId", "prefix", "name", "type", "sourceNavigatorItemId", "children", )

    def __init__(self, navigatorItemId: NavigatorItemId, prefix: str, name: str, type: str, sourceNavigatorItemId: Optional[NavigatorItemId] = None, children: Optional[List[NavigatorItemArrayItem]] = None):
        self.navigatorItemId: NavigatorItemId = navigatorItemId
        self.prefix: str = prefix
        self.name: str = name
        self.type: str = type
        self.sourceNavigatorItemId: Optional[NavigatorItemId] = sourceNavigatorItemId
        self.children: Optional[List[NavigatorItemArrayItem]] = children

NavigatorItem.get_classinfo().add_field('navigatorItemId', NavigatorItemId)
NavigatorItem.get_classinfo().add_field('prefix', str)
NavigatorItem.get_classinfo().add_field('name', str)
NavigatorItem.get_classinfo().add_field('type', str, value_set(['UndefinedItem', 'ProjectMapRootItem', 'StoryItem', 'SectionItem', 'ElevationItem', 'InteriorElevationItem', 'WorksheetItem', 'DetailItem', 'DocumentFrom3DItem', 'Perspective3DItem', 'Axonometry3DItem', 'CameraSetItem', 'CameraItem', 'ScheduleItem', 'ProjectIndexItem', 'TextListItem', 'GraphicListItem', 'InfoItem', 'HelpItem', 'FolderItem', 'LayoutBookRootItem', 'SubsetItem', 'LayoutItem', 'DrawingItem', 'MasterFolderItem', 'MasterLayoutItem']))
NavigatorItem.get_classinfo().add_field('sourceNavigatorItemId', Optional[NavigatorItemId])
NavigatorItem.get_classinfo().add_field('children', Optional[List[NavigatorItemArrayItem]])

NavigatorItem_ = NavigatorItem
NavigatorItemArrayItem.get_classinfo().add_field('navigatorItem', NavigatorItem)


class NavigatorTree(_ACBaseType):
    """ A tree of navigator items.

    Attributes:
        rootItem (:obj:`NavigatorItem`): Details of a navigator item.

    """
    __slots__ = ("rootItem", )

    def __init__(self, rootItem: NavigatorItem):
        self.rootItem: NavigatorItem = rootItem

NavigatorTree.get_classinfo().add_field('rootItem', NavigatorItem)


class Types:
    """ 
    """
    ClassificationSystemId=ClassificationSystemId
    ClassificationSystemIdArrayItem=ClassificationSystemIdArrayItem
    ClassificationItemId=ClassificationItemId
    ClassificationItemIdArrayItem=ClassificationItemIdArrayItem
    ClassificationId=ClassificationId
    ClassificationItemDetails=ClassificationItemDetails
    ClassificationSystem=ClassificationSystem
    UserDefinedPropertyUserId=UserDefinedPropertyUserId
    BuiltInPropertyUserId=BuiltInPropertyUserId
    PropertyUserId=PropertyUserId
    PropertyId=PropertyId
    PropertyIdArrayItem=PropertyIdArrayItem
    PropertyGroup=PropertyGroup
    NormalNumberPropertyValue=NormalNumberPropertyValue
    NormalIntegerPropertyValue=NormalIntegerPropertyValue
    NormalStringPropertyValue=NormalStringPropertyValue
    NormalBooleanPropertyValue=NormalBooleanPropertyValue
    NormalLengthPropertyValue=NormalLengthPropertyValue
    NormalAreaPropertyValue=NormalAreaPropertyValue
    NormalVolumePropertyValue=NormalVolumePropertyValue
    NormalAnglePropertyValue=NormalAnglePropertyValue
    NormalNumberListPropertyValue=NormalNumberListPropertyValue
    NormalIntegerListPropertyValue=NormalIntegerListPropertyValue
    NormalStringListPropertyValue=NormalStringListPropertyValue
    NormalBooleanListPropertyValue=NormalBooleanListPropertyValue
    NormalLengthListPropertyValue=NormalLengthListPropertyValue
    NormalAreaListPropertyValue=NormalAreaListPropertyValue
    NormalVolumeListPropertyValue=NormalVolumeListPropertyValue
    NormalAngleListPropertyValue=NormalAngleListPropertyValue
    UserUndefinedPropertyValue=UserUndefinedPropertyValue
    NotAvailablePropertyValue=NotAvailablePropertyValue
    NotEvaluatedPropertyValue=NotEvaluatedPropertyValue
    DisplayValueEnumId=DisplayValueEnumId
    EnumValueId=EnumValueId
    PossibleEnumValue=PossibleEnumValue
    PossibleEnumValuesArrayItem=PossibleEnumValuesArrayItem
    Error=Error
    ErrorItem=ErrorItem
    SuccessfulExecutionResult=SuccessfulExecutionResult
    FailedExecutionResult=FailedExecutionResult
    ExecutionResult=ExecutionResult
    ElementId=ElementId
    ElementIdArrayItem=ElementIdArrayItem
    ElementsWrapper=ElementsWrapper
    ElementsOrError=ElementsOrError
    NavigatorItemId=NavigatorItemId
    PublisherSetId=PublisherSetId
    OtherNavigatorTreeId=OtherNavigatorTreeId
    FolderParameters=FolderParameters
    NavigatorTreeId=NavigatorTreeId
    BoundingBox2D=BoundingBox2D
    BoundingBox3D=BoundingBox3D
    Subset=Subset
    LayoutParameters=LayoutParameters
    ClassificationIdWrapper=ClassificationIdWrapper
    ClassificationItemDetailsWrapper=ClassificationItemDetailsWrapper
    EnumValueIdWrapper=EnumValueIdWrapper
    NavigatorItemIdWrapper=NavigatorItemIdWrapper
    BoundingBox2DWrapper=BoundingBox2DWrapper
    BoundingBox3DWrapper=BoundingBox3DWrapper
    ClassificationIdOrError=ClassificationIdOrError
    ElementClassification=ElementClassification
    ClassificationItemOrError=ClassificationItemOrError
    PropertyIdOrError=PropertyIdOrError
    PropertyDefinition=PropertyDefinition
    NormalSingleEnumPropertyValue=NormalSingleEnumPropertyValue
    NormalMultiEnumPropertyValue=NormalMultiEnumPropertyValue
    BoundingBox2DOrError=BoundingBox2DOrError
    BoundingBox3DOrError=BoundingBox3DOrError
    ClassificationIdsOrErrorsWrapper=ClassificationIdsOrErrorsWrapper
    PropertyDefinitionWrapper=PropertyDefinitionWrapper
    ElementClassificationOrError=ElementClassificationOrError
    PropertyDefinitionOrError=PropertyDefinitionOrError
    ClassificationItemArrayItem=ClassificationItemArrayItem
    ClassificationItemInTree=ClassificationItemInTree
    NormalOrUserUndefinedPropertyValue=NormalOrUserUndefinedPropertyValue
    ElementPropertyValue=ElementPropertyValue
    PropertyValue=PropertyValue
    PropertyValueWrapper=PropertyValueWrapper
    PropertyValueOrErrorItem=PropertyValueOrErrorItem
    PropertyValuesWrapper=PropertyValuesWrapper
    PropertyValuesOrError=PropertyValuesOrError
    NavigatorItemArrayItem=NavigatorItemArrayItem
    NavigatorItem=NavigatorItem
    NavigatorTree=NavigatorTree




import importlib
import os
from typing import Tuple
from .releases import Types, Commands, Utilities

class _Versioning:
    __slots__ = ('types', 'commands', 'utilities')

    @staticmethod
    def discover(release_: int, build_: int) -> Tuple[int, int]:
        import archicad.releases as RELS
        all_releases = sorted((int(dir_entry.name.lstrip('ac')) for dir_entry in os.scandir(RELS.__path__[0])
            if dir_entry.is_dir() and dir_entry.name.startswith('ac')), reverse=True)
        release = next(x for x in all_releases if x <= release_)

        BUILDS = importlib.import_module(f'archicad.releases.ac{release}')
        all_builds = sorted({int(dir_entry.name.strip('btypescomanduil.')) for dir_entry in os.scandir(BUILDS.__path__[0])
            if dir_entry.is_file() and dir_entry.name.startswith('b')}, reverse=True)
        build = next(x for x in all_builds if x <= build_)
        return release, build

    def __init__(self, release: int, build: int, request):
        assert release >= 24
        self.types = Types
        self.commands = Commands
        self.utilities = Utilities
        self.load(*_Versioning.discover(release, build), request)
    
    def load(self, release: int, build: int, request):
        self.types = importlib.import_module(f'archicad.releases.ac{release}.b{build}types').Types()
        self.commands = importlib.import_module(f'archicad.releases.ac{release}.b{build}commands').Commands(request)
        self.utilities = importlib.import_module(f'archicad.releases.ac{release}.b{build}utilities').Utilities(self.types, self.commands)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="archicad",
    version="24.3000",
    author="GRAPHISOFT SE",
    author_email="archicadapi@graphisoft.com",
    description="Python binding for the ARCHICAD JSON command interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://archicadapi.graphisoft.com/archicadPythonPackage",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    license='Apache'
)
